# Type stubs for cyminhook

class MinHook:
    def __init__(self, *, signature=None, target=None, detour=None):
        ...

    def close(self):
        """Close the hook. Removing it."""

    def enable(self):
        """Enable the hook."""

    def disable(self):
        """Disable the hook."""

def queue_enable(hook: MinHook) -> None:
    """Queue to enable an already created hook."""

def queue_disable(hook: MinHook) -> None:
    """Queue to disable an already created hook."""

def apply_queued() -> None:
    """Apply all queued changes in one go."""



from enum import IntEnum

import pytest

from pymhf.extensions.ctypes import c_enum32


def test_invalid_c_enum32_cases():
    with pytest.raises(TypeError):
        c_enum32[22]


def test_c_enum_members():
    class Alphabet(IntEnum):
        A = 0
        B = 1
        C = 2
        D = 3
        E = 4

    assert c_enum32[Alphabet]._members() == ["A", "B", "C", "D", "E"]



import subprocess
from multiprocessing.connection import Connection
from typing import Optional

import psutil


class stdoutSocket:
    def __init__(self, connection: Connection):
        self.connection = connection

    def write(self, val):
        self.connection.send_bytes(val.encode())

    def flush(self):
        pass


def open_log_console(log_script: str, log_dir: str, name_override: str = "pymhf console") -> Optional[int]:
    """Open the logging console and return the pid of it.

    Parameters
    ----------
    log_script
        Path to the logging script to be run.
    log_dir:
        Path where the log files will be written to.
    name_override:
        Name to override the default "pymhf console" window title.
    """
    cmd = f'''cmd.exe /c "start "{name_override}" python "{log_script}" "{log_dir}""'''
    with subprocess.Popen(cmd) as proc:
        log_ppid = proc.pid
    for proc in psutil.process_iter(["pid", "name", "ppid"]):
        if proc.info["ppid"] == log_ppid:
            return proc.info["pid"]



import ctypes
import typing
from ctypes import _Pointer

from ._types import DetourTime  # noqa

# Monkeypatch typed _Pointer from typeshed into ctypes.
# c/o https://github.com/python/mypy/issues/7540#issuecomment-845741357
if not typing.TYPE_CHECKING:
    ctypes._Pointer_orig = _Pointer

    class _Pointer(ctypes._Pointer):
        def __class_getitem__(cls, item):
            return ctypes.POINTER(item)

    ctypes._Pointer = _Pointer


# TODO: Assess whether we need this any more...
def pymhf_overload(func):
    setattr(func, "_is_overloaded", True)
    return func


_typing_overload = typing.overload  # noqa
typing.overload = pymhf_overload



import ctypes
from concurrent.futures import ThreadPoolExecutor

from pymhf.core._types import LoadTypeEnum, pymhfConfig

CWD: str = ""
MODULE_PATH: str = ""
HANDLE = None
MAIN_HWND = None
DPG_HWND = None
PID: int = -1
BINARY_HASH: str = ""
BASE_ADDRESS: int = -1
SIZE_OF_IMAGE: int = -1
CONFIG: pymhfConfig = {}
EXE_NAME: str = ""
BINARY_PATH: str = ""
LOAD_TYPE: LoadTypeEnum = LoadTypeEnum.INVALID
MOD_SAVE_DIR: str = ""
INCLUDED_ASSEMBLIES: dict[str, str] = {}
CACHE_DIR: str = ""
_SENTINEL_PTR: int = 0

_executor: ThreadPoolExecutor = None  # type: ignore

imports: dict[str, dict[str, ctypes._CFuncPtr]] = {}



import argparse
import os
import os.path as op
import shutil
import subprocess
import sys
from enum import Enum
from importlib.metadata import PackageNotFoundError, entry_points, version
from typing import Optional

import questionary

from .core._types import FUNCDEF  # noqa
from .core.hooking import FuncHook  # noqa
from .core.mod_loader import Mod, ModState  # noqa
from .main import load_mod_file, load_module, run_module  # noqa
from .utils.imports import SPHINX_AUTODOC_RUNNING

try:
    __version__ = version("pymhf")
except PackageNotFoundError:
    pass

try:
    from tkinter import Tk, filedialog

    has_tkinter = True
except ModuleNotFoundError:
    has_tkinter = False


class ModeEnum(Enum):
    INVALID = 0
    RUN = 1
    CONFIG = 2
    INIT = 3
    COMMAND = 4
    # DATA = 5


class FolderModeEnum(Enum):
    INVALID = 0
    LIBRARY = 1
    MOD_FOLDER = 2
    NON_LOCAL = 3


def _is_int(val: str) -> bool:
    try:
        int(val)
    except (ValueError, TypeError):
        return False
    return True


def get_folder(title: str, q: questionary.Question, has_tkinter: bool, idir: Optional[str] = None) -> str:
    if has_tkinter:
        return filedialog.askdirectory(initialdir=idir, title=title)
    else:
        return q.ask()


# Pattern shamelessly stolen from the questionary library (which is the actual reason this code has issues in
# the first place!)
if not SPHINX_AUTODOC_RUNNING and os.environ.get("PYTEST_VERSION") is None:
    START_PAUSED = questionary.confirm("Start the game paused?", default=True)
    RUN_GAME = questionary.confirm("Run game?", default=True)
    STEAM_ID_Q = questionary.text("Enter the steam game ID:", validate=_is_int)
    EXE_PATH_Q = questionary.path("Enter the absolute path to the binary:")
    MOD_DIR_Q = questionary.path("Enter the absolute path the mod directory")
    MOD_SAVE_DIR_Q = questionary.path("Enter the absolute path the mod save directory")
    LOG_DIR_Q = questionary.path("Enter the absolute path the logs directory")
    CONTINUE_CONFIGURING_Q = questionary.confirm("Would you like to configure more options?", default=True)

    CFG_OPT_BIN_PATH = "Set binary path"
    CFG_OPT_MOD_PATH = "Set mod directory"
    CFG_OPT_MOD_SAVE_PATH = "Set mod save directory"
    CFG_OPT_STEAM_ID = "Configure steam game id"
    CFG_OPT_LOG_PATH = "Set log directory"
    CFG_OPT_START_PAUSED = "Set game to start paused"
    CONFIG_SELECT_Q = questionary.select(
        "What would you like to configure?",
        choices=[
            CFG_OPT_BIN_PATH,
            CFG_OPT_MOD_PATH,
            CFG_OPT_STEAM_ID,
            CFG_OPT_LOG_PATH,
            CFG_OPT_MOD_SAVE_PATH,
            CFG_OPT_START_PAUSED,
        ],
    )

# This is the name of the config file within the library.
CFG_FILENAME = "pymhf.toml"
# This is the name of the config file which will be sorted in the user data folder.
LOCAL_CFG_FILENAME = "pymhf.local.toml"


def run():
    """Main entrypoint which can be used to run programs with pymhf.
    This will take the first argument as the name of a module which has been installed.
    """
    from .utils.parse_toml import _parse_toml, read_pymhf_settings, write_pymhf_settings

    parser = argparse.ArgumentParser(
        prog="pymhf",
        description="Run the registered plugin",
    )

    parser.add_argument("--version", action="store_true", help="Print the version of pyMHF and exit")

    command_parser = parser.add_subparsers(dest="_command")

    # `run` command parser
    run_parser = command_parser.add_parser(
        "run",
        help=(
            "Run the provided plugin name (if registered), or the provided path to a module or single-file "
            "mod."
        ),
    )
    run_parser.add_argument(
        "plugin_name",
        help=(
            "The name of the installed library to run, or the single-file script to run, or the path to a "
            "folder which contains a library to run locally."
        ),
    )

    config_parser = command_parser.add_parser(
        "config",
        help=(
            "Configure the provided plugin name (if registered), or the provided path to a module or "
            "single-file mod."
        ),
    )
    config_parser.add_argument(
        "-o",
        "--open",
        action="store_true",
        default=False,
        help="Open the directory the local config is stored at.",
    )
    config_parser.add_argument(
        "plugin_name",
        help=(
            "The name of the installed library to run, or the single-file script to run, or the path to a "
            "folder which contains a library to run locally."
        ),
    )

    init_parser = command_parser.add_parser(
        "init", help="[NOT YET IMPLEMENTED] Initialise a new empty project under the given name."
    )
    init_parser.add_argument(
        "plugin_name",
        help=(
            "The name of the installed library to run, or the single-file script to run, or the path to a "
            "folder which contains a library to run locally."
        ),
    )

    cmd_parser = command_parser.add_parser("cmd", help="Perform a registered command for the library")
    cmd_parser.add_argument(
        "command",
        help=(
            "The command to be run by the specified library. These are registered in the "
            "[project.entry-points.pymhfcmd] section of the pyproject.toml file."
        ),
    )
    cmd_parser.add_argument(
        "plugin_name",
        help=(
            "The name of the installed library to run, or the path to a folder which contains a library to "
            "run locally."
        ),
    )

    # data_parser = command_parser.add_parser(
    #     "data", help="Configure the data for the provided plugin name (if registered), or the provided path"
    #     " to a module."
    # )
    # data_parser.add_argument(
    #     "plugin_name",
    #     help=(
    #         "The name of the installed library to run, or the path to a folder which contains a library to "
    #         "run locally."
    #     ),
    # )
    # data_parser.add_argument(
    #     "-H",
    #     "--hash",
    #     action="store_true",
    #     help="Get the current binary hash",
    # )

    args, extras = parser.parse_known_args()  # noqa

    if args.version:
        print(__version__)
        sys.exit(0)

    # TODO: The extras can be passed to the registered library in the future.

    try:
        plugin_name: str = args.plugin_name
    except AttributeError:
        parser.error("No command provided. Please provide a valid command. See `pymhf --help` for details.")
    command = args._command
    if command == "run":
        mode = ModeEnum.RUN
    elif command == "config":
        mode = ModeEnum.CONFIG
    elif command == "init":
        mode = ModeEnum.INIT
    elif command == "cmd":
        mode = ModeEnum.COMMAND
    # elif command == "data":
    #     mode = ModeEnum.DATA
    else:
        mode = ModeEnum.INVALID
    standalone = False
    local = False

    if op.isfile(plugin_name) and op.exists(plugin_name):
        # In this case we are running in stand-alone mode
        standalone = True
        if plugin_name.endswith(".py"):
            plugin_name = op.realpath(plugin_name)

    if op.isdir(plugin_name) and op.exists(plugin_name):
        # In this case we are running a library directly from pymhf.
        # This can be done for two reasons... We either want to actually run it, or we are configuring it.
        local = True

    if standalone:
        if mode != ModeEnum.RUN:
            print("Only the `run` command can be used with single-file mods.")
            if mode == ModeEnum.CONFIG:
                print(
                    "To configure a single-file mod, change the values directly in the inline script "
                    "metadata. See https://monkeyman192.github.io/pyMHF/docs/settings.html"
                )
            return
        load_mod_file(plugin_name)
        return

    cmd_command = None
    if mode == ModeEnum.COMMAND:
        cmd_command = args.command

    folder_mode = FolderModeEnum.INVALID

    if local:
        # Check to see if the folder we are pointing to has a pyproject.toml file, or a pymhf.toml file.
        # If it's a pyproject.toml, then we are running a library directly.
        # If it's a pymhf.toml, then it's a mod folder.
        # Parse the pyproject.toml file to get some info...
        # We'll give precedency to the pyproject.toml file.
        local_plugin_dir = op.realpath(plugin_name)
        pyproject_toml = op.join(local_plugin_dir, "pyproject.toml")
        pymhf_toml = op.join(local_plugin_dir, CFG_FILENAME)
        if op.exists(pyproject_toml):
            folder_mode = FolderModeEnum.LIBRARY
        elif op.exists(pymhf_toml):
            folder_mode = FolderModeEnum.MOD_FOLDER

        if folder_mode == FolderModeEnum.INVALID:
            print(
                f"Error: No pyproject.toml or pymhf.toml file in the directory {local_plugin_dir}."
                "Please ensure the target project has one of these."
            )
            return
        if folder_mode == FolderModeEnum.LIBRARY:
            settings = _parse_toml(pyproject_toml, False)
            plugin_name = None
            if (
                project_cfg := settings.get("project", {}).get("entry-points", {}).get("pymhflib")
            ) is not None:
                plugin_name = list(project_cfg.keys())[0]
                if mode == ModeEnum.COMMAND:
                    if (
                        command_cfg := settings.get("project", {}).get("entry-points", {}).get("pymhfcmd")
                    ) is not None:
                        from .core.importing import import_file

                        if (resolved_command := command_cfg.get(cmd_command)) is not None:
                            mod = import_file(op.join(local_plugin_dir, plugin_name))
                            cmd = getattr(mod, resolved_command, None)
                            if cmd is not None:
                                cmd(extras)
                                return
                            else:
                                print(
                                    f"Command {resolved_command!r} cannot be found in module {plugin_name}. "
                                    "Please ensure the function exists and can be run from this module."
                                )
                                return
                        else:
                            print(f"No registered command {cmd_command!r}. Please check your configuration.")
                            return
                    else:
                        print(
                            "No [project.entry-points.pymhfcmd] section found in the pyproject.toml. "
                            "Please create one to register and run custom commands."
                        )
                        return
        elif folder_mode == FolderModeEnum.MOD_FOLDER:
            # Parse the pymhf.toml file to get some details about what we are running.
            settings = _parse_toml(pymhf_toml, False)
            # We'll set the plugin name as the name of the folder
            plugin_name = op.basename(local_plugin_dir)
        if plugin_name is None:
            print(
                f"Cannot determine the project at the path specified {plugin_name}.\n"
                "Please ensure the pyproject.toml file has the [project.entry-points.pymhflib] entry."
                "This process will not continue."
            )
            return
    else:
        folder_mode = FolderModeEnum.NON_LOCAL

    # Get the location of app data, then construct the expected folder name.
    appdata_data = os.environ.get("APPDATA", op.expanduser("~"))
    if appdata_data == "~":
        # In this case the APPDATA environment variable isn't set and ~ also fails to resolve.
        # Raise a error and stop.
        print("Critical Error: Cannot find user directory. Ensure APPDATA environment variable is set")
        exit()
    cfg_folder = op.join(appdata_data, "pymhf", plugin_name)
    os.makedirs(cfg_folder, exist_ok=True)

    initial_config = False

    if not local:
        eps = entry_points()
        # This check is to ensure compatibility with multiple versions of python as the code 3.10+ isn't
        # backward compatible.
        if isinstance(eps, dict):
            pymhf_entry_points = eps.get("pymhflib", [])
        else:
            pymhf_entry_points = eps.select(group="pymhflib")
        if mode == ModeEnum.COMMAND:
            if isinstance(eps, dict):
                pymhf_commands = eps.get("pymhfcmd", [])
            else:
                pymhf_commands = eps.select(group="pymhfcmd")

            if not pymhf_commands:
                print(
                    "No [project.entry-points.pymhfcmd] section found in the pyproject.toml. "
                    "Please create one to register and run custom commands."
                )

        required_lib = None
        resolved_command = None

        for ep in pymhf_entry_points:
            if ep.name.lower() == plugin_name.lower():
                required_lib = ep
        if mode == ModeEnum.COMMAND:
            for cmd in pymhf_commands:
                # See if the library has a check command defined.
                if cmd.name.lower() == cmd_command:
                    resolved_command = cmd.value

        if required_lib is None:
            print(
                f"Cannot find {plugin_name} as an installed plugin. "
                "Please ensure it has been installed and try again"
            )
            return
        folder_mode = FolderModeEnum.LIBRARY

        module_dir = op.dirname(required_lib.load().__file__)

        if mode == ModeEnum.COMMAND:
            if resolved_command is not None:
                # We can use importlib here since the library has to be installed if we have made it here.
                import importlib

                mod = importlib.import_module(plugin_name)
                cmd = getattr(mod, resolved_command, None)
                if cmd is not None:
                    cmd(extras)
                    return
                else:
                    print(
                        f"Command {resolved_command!r} cannot be found in module {plugin_name}. "
                        "Please ensure the function exists and can be run from this module."
                    )
                    return
            else:
                print(f"No registered command {cmd_command!r}. Please check your configuration.")
                return

    else:
        if folder_mode == FolderModeEnum.LIBRARY:
            module_dir = op.join(local_plugin_dir, plugin_name)
        else:
            module_dir = local_plugin_dir

    cfg_file = op.join(module_dir, CFG_FILENAME)
    config_progress_file = op.join(cfg_folder, ".config_in_progress")
    if not op.exists(cfg_file):
        print(
            f"Cannot find `{CFG_FILENAME}` for {plugin_name}! This is likely an error on the maintainers' "
            "behalf.\nCannot continue loading until this is fixed."
        )
        return
    else:
        dst = op.join(cfg_folder, LOCAL_CFG_FILENAME)
        if not op.exists(dst) or op.exists(config_progress_file):
            # In this case we can prompt the user to enter the config values which need to be changed.
            initial_config = True

    if initial_config:
        local_config = {"local_config": {}}
        if not has_tkinter:
            print(
                "tkinter cannot be found. Please ensure it's installed as part of your python install.\n"
                "Falling back to the command-line method of providing an input."
            )
        # Copy the config file to the appdata directory.
        shutil.copyfile(cfg_file, dst)
        # Write the file which indicates we are in progress.
        with open(config_progress_file, "w") as f:
            f.write("")

        if folder_mode != FolderModeEnum.MOD_FOLDER:
            # Modify some of the values in the config file, allowing the user to enter the values they want.
            # We only do this if we are not dealing with a mod folder...
            if has_tkinter:
                root = Tk()
                root.withdraw()
            if (
                mod_folder := get_folder("Select folder where mods are located", MOD_DIR_Q, has_tkinter)
            ) is not None:
                local_config["local_config"]["mod_dir"] = mod_folder
            else:
                return None

        # Write the config back and then delete the temporary file only once everything is ok.
        write_pymhf_settings(local_config, dst)
        os.remove(config_progress_file)
        initial_config = False
    elif mode == ModeEnum.CONFIG:
        if args.open:
            print(f"Opening {cfg_folder!r} and exiting")
            subprocess.call(f'explorer "{cfg_folder}"')
            return
        local_config = read_pymhf_settings(dst)
        pymhf_settings = local_config["local_config"]
        keep_going = True
        if has_tkinter:
            root = Tk()
            root.withdraw()
        while keep_going:
            config_choice = CONFIG_SELECT_Q.ask()
            if config_choice == CFG_OPT_BIN_PATH:
                if (exe_path := EXE_PATH_Q.ask()) is not None:
                    pymhf_settings["exe_path"] = exe_path
                    if "steam_gameid" in pymhf_settings:
                        del pymhf_settings["steam_gameid"]
                else:
                    return
            elif config_choice == CFG_OPT_STEAM_ID:
                if (steam_id := STEAM_ID_Q.ask()) is not None:
                    pymhf_settings["steam_gameid"] = steam_id
                    if "exe_path" in pymhf_settings:
                        del pymhf_settings["exe_path"]
                else:
                    return
            elif config_choice == CFG_OPT_MOD_PATH:
                if (
                    mod_dir := get_folder("Select folder where mods can be found", MOD_DIR_Q, has_tkinter)
                ) is not None:
                    pymhf_settings["mod_dir"] = mod_dir
                else:
                    return None
            elif config_choice == CFG_OPT_START_PAUSED:
                if (start_paused := START_PAUSED.ask()) is not None:
                    pymhf_settings["start_paused"] = start_paused
                else:
                    return
            elif config_choice == CFG_OPT_LOG_PATH:
                if (
                    log_dir := get_folder("Select folder where logs are placed", LOG_DIR_Q, has_tkinter)
                ) is not None:
                    pymhf_settings["log_dir"] = log_dir
                else:
                    return None
            elif config_choice == CFG_OPT_MOD_SAVE_PATH:
                if (
                    msd := get_folder("Select folder where mod saves are placed", MOD_SAVE_DIR_Q, has_tkinter)
                ) is not None:
                    pymhf_settings["mod_save_dir"] = msd
                else:
                    return None
            elif config_choice is None:
                keep_going = False

            keep_going = CONTINUE_CONFIGURING_Q.ask()
            if keep_going is None:
                return
        local_config["local_config"] = pymhf_settings
        write_pymhf_settings(local_config, dst)

        if not RUN_GAME.ask():
            return
    # Final step: If we have either finished the initial config, or we didn't want to config at all, run the
    # binary.
    if not initial_config:
        load_module(plugin_name, module_dir)



# /// script
# dependencies = ["pymhf[gui]>=0.1.15"]
# 
# [tool.pymhf]
# exe = "notepad.exe"
# start_paused = false
# start_exe = false
# interactive_console = false
#
# [tool.pymhf.gui]
# always_on_top = false
# 
# [tool.pymhf.logging]
# shown = false
# log_dir = "."
# log_level = "info"
# window_name_override = "pyMHF Test Mod"
# ///
from logging import getLogger

from pymhf import Mod, load_mod_file

logger = getLogger("TestMod")


class TestMod(Mod):
    __author__ = "monkeyman192"
    __description__ = "Test mod to check if pyMHF works"

    def __init__(self):
        super().__init__()
        logger.info("Loaded Test mod!")


if __name__ == "__main__":
    load_mod_file(__file__)



import ast
import importlib
import importlib.util
import logging
import os.path as op
import string
import sys
from types import ModuleType
from typing import Optional

logger = logging.getLogger(__name__)


VALID_CHARS = string.ascii_letters + string.digits + "_"


def _clean_name(name: str) -> str:
    """Remove any disallowed characters from the filename so that we get a
    valid module name.
    """
    out = ""
    for char in name:
        if char not in VALID_CHARS:
            out += "_"
        else:
            out += char
    return out


def _fully_unpack_ast_attr(obj: ast.Attribute) -> str:
    name = ""
    _obj = obj
    while isinstance(_obj, ast.Attribute):
        if name:
            name = f"{_obj.attr}.{name}"
        else:
            name = _obj.attr
        _obj = _obj.value
    else:
        if isinstance(_obj, ast.Name):
            name = f"{_obj.id}.{name}"
    return name


def parse_file_for_mod(data: str) -> bool:
    """Parse the provided data and determine if there is at least one mod class in it."""
    tree = ast.parse(data)
    mod_class_name = None
    for node in tree.body:
        # First, determine the name the Mod object is imported as.
        if isinstance(node, ast.Import):
            for node_ in node.names:
                if isinstance(node_, ast.alias):
                    if node_.name in ("pymhf", "pymhf.core.mod_loader"):
                        mod_class_name = (node_.asname or node_.name) + ".Mod"
        if isinstance(node, ast.ImportFrom):
            if node.module in ("pymhf", "pymhf.core.mod_loader"):
                for node_ in node.names:
                    if isinstance(node_, ast.alias):
                        if node_.name == "Mod":
                            mod_class_name = node_.asname or node_.name
        # Now, when we go over the class nodes, check the base classes.
        if isinstance(node, ast.ClassDef):
            for base in node.bases:
                # For a simple name, it's easy - just match it.
                if isinstance(base, ast.Name):
                    if base.id == mod_class_name:
                        return True
                # If it's an attribute it's a bit trickier...
                elif isinstance(base, ast.Attribute):
                    resolved_base = _fully_unpack_ast_attr(base)
                    if resolved_base == mod_class_name:
                        return True
    return False


def import_file(fpath: str) -> Optional[ModuleType]:
    try:
        module_name = _clean_name(op.splitext(op.basename(fpath))[0])
        if op.isdir(fpath):
            # If a directory is passed in, then add __init__.py to it so that
            # we may correctly import it.
            fpath = op.join(fpath, "__init__.py")
        if spec := importlib.util.spec_from_file_location(module_name, fpath):
            module = importlib.util.module_from_spec(spec)
            module.__name__ = module_name
            module.__spec__ = spec
            sys.modules[module_name] = module
            if spec.loader:
                spec.loader.exec_module(module)
                return module
        else:
            print("failed")
    except Exception:
        logger.exception(f"Error loading {fpath}")



from enum import Enum
from typing import Any, Protocol


class VariableType(Enum):
    INTEGER = 0
    FLOAT = 1
    STRING = 2
    BOOLEAN = 3


class ButtonProtocol(Protocol):
    _is_button: bool
    _button_text: str

    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...


class ComboBoxProtocol(Protocol):
    _is_combobox: bool
    _combobox_text: str
    _items: list[str]

    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...


class VariableProtocol(Protocol):
    _is_variable: bool
    _variable_type: VariableType
    _label_text: str
    _has_setter: bool
    _extra_args: dict

    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...



from enum import Enum
from typing import Any, NamedTuple, Optional, Protocol

from typing_extensions import NotRequired, TypedDict


class pymhfGUIConfig(TypedDict):
    shown: NotRequired[bool]
    scale: NotRequired[float]
    always_on_top: NotRequired[bool]


class pymhfLoggingConfig(TypedDict):
    default_log_dir: NotRequired[str]
    log_level: NotRequired[str]
    window_name_override: NotRequired[str]
    shown: NotRequired[bool]


class pymhfConfig(TypedDict):
    exe: NotRequired[str]
    pid: NotRequired[int]
    steam_guid: NotRequired[int]
    args: NotRequired[list[str]]
    required_assemblies: NotRequired[list[str]]
    start_paused: NotRequired[bool]
    default_mod_save_dir: NotRequired[str]
    internal_mod_dir: NotRequired[str]
    start_exe: NotRequired[bool]
    interactive_console: NotRequired[bool]
    logging: NotRequired[pymhfLoggingConfig]
    gui: NotRequired[pymhfGUIConfig]


class LoadTypeEnum(Enum):
    INVALID = 0
    SINGLE_FILE = 1
    LIBRARY = 2
    MOD_FOLDER = 3


class FunctionIdentifier(NamedTuple):
    name: str
    offset: int
    binary: str
    is_absolute: bool

    def __eq__(self, other: "FunctionIdentifier"):
        """Two FunctionIdentifier's are the same if their offset and binary are the same."""
        return self.offset == other.offset and self.binary == other.binary

    def __hash__(self):
        """Only use the offset and binary values for the hash. Name doesn't matter."""
        return hash((self.offset, self.binary))


# TODO: Fully deprecate.
class FUNCDEF(NamedTuple):
    restype: Any
    argtypes: list


class DetourTime(Enum):
    NONE = 0
    BEFORE = 1
    AFTER = 2


class KeyPressProtocol(Protocol):
    _hotkey: str
    _hotkey_press: str


class CustomTriggerProtocol(Protocol):
    _custom_trigger: Optional[str]
    _description: Optional[str]
    _hook_time: DetourTime

    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...


class HookProtocol(Protocol):
    _disabled: bool
    _is_funchook: bool
    _is_manual_hook: bool
    _is_imported_func_hook: bool
    _is_exported_func_hook: bool
    _has__result_: bool
    _hook_func_name: str
    _hook_func_def: FUNCDEF
    _hook_offset: Optional[int]
    _hook_pattern: Optional[str]
    _hook_time: DetourTime
    _func_overload: Optional[str]
    _get_caller: Optional[bool]
    _noop: Optional[bool]
    _dll_name: Optional[str]

    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...


class CallerHookProtocol(HookProtocol):
    def caller_address(self) -> int:
        """The address relative to the base of the calling binary that this functiuon detour was called from.
        Note that this will be the address one instruction after the caller since this value is where
        execution returns to after the detours and original function have been run.
        """
        ...


# TODO: Remove once we have got the binary able to be specified for function_hook decorators.
class ManualHookProtocol(HookProtocol):
    _hook_binary: Optional[str]



from ctypes import CFUNCTYPE, WinDLL
from logging import getLogger
from typing import Any, Optional

import pymhf.core._internal as _internal
from pymhf.core._types import FUNCDEF
from pymhf.core.errors import UnknownFunctionError
from pymhf.core.memutils import _get_binary_info, find_pattern_in_binary
from pymhf.core.module_data import module_data
from pymhf.core.utils import saferun_decorator

calling_logger = getLogger("CallingManager")


# TODO: Everything in this file is deprecated. DO NOT use it.


@saferun_decorator
def call_exported(name: str, func_def: FUNCDEF, *args):
    """Call a function exported by the main binary.

    Parameters
    ----------
    name:
        The name of the exported function. This will generally be the mangled name of the function as provided
        by IDA/ghidra.
    func_def:
        The restype and argtypes of the function being called.
    args:
        The arguments to be passed to the function being called.
    """
    # TODO: Improve this so that the own_dll is cached as well as the func_def so that we only incur a slight
    # performance hit the very first time.
    own_dll = WinDLL(_internal.BINARY_PATH)
    func_ptr = getattr(own_dll, name)
    func_ptr.restype = func_def.restype
    func_ptr.argtypes = func_def.argtypes
    return func_ptr(*args)


@saferun_decorator
def call_function(
    name: str,
    *args,
    overload: Optional[str] = None,
    pattern: Optional[str] = None,
    func_def: Optional[FUNCDEF] = None,
    offset: Optional[int] = None,
    binary: Optional[str] = None,
) -> Any:
    """Call a named function.

    Parameters
    ----------
    name
        The name of the function to be called.
        For now the function signature will be looked up from the known signatures by name.
    args
        The args to pass to the function call.
    overload
        The overload name to be called if required.
    pattern
        The pattern which can be used to find where the function is.
        If provided this will be used instead of the offset as determined by the name.
    offset
        The offset relative to the binary.
        If provided it will take precedence over the `pattern` argument.
    binary
        The name of the binary to search for the pattern within, or to find the offset relative to.
        If not provided, will fallback to the name of the binary as provided by the `exe` config value.
    """
    if func_def is not None:
        _sig = func_def
    else:
        _sig = module_data.FUNC_CALL_SIGS[name]
    # If we have a binary defined in module_data, use it.
    binary = binary or module_data.FUNC_BINARY
    if offset is None:
        if pattern:
            offset = find_pattern_in_binary(pattern, False, binary)
        else:
            if (_pattern := module_data.FUNC_PATTERNS.get(name)) is not None:
                if isinstance(_pattern, str):
                    offset = find_pattern_in_binary(_pattern, False, binary)
                else:
                    if (opattern := _pattern.get(overload)) is not None:
                        offset = find_pattern_in_binary(opattern, False, binary)
                    else:
                        first = list(_pattern.items())[0]
                        calling_logger.warning(f"No pattern overload was provided for {name}. ")
                        calling_logger.warning(f"Falling back to the first overload ({first[0]})")
                        offset = find_pattern_in_binary(first[1], False, binary)
            else:
                offset = module_data.FUNC_OFFSETS.get(name)
            if offset is None:
                raise UnknownFunctionError(f"Cannot find function {name}")

    if isinstance(_sig, FUNCDEF):
        sig = CFUNCTYPE(_sig.restype, *_sig.argtypes)
    elif isinstance(_sig, dict):
        # TODO: Check to see if _sig is actually a dict. If it's not then we should raise an error.
        # Look up the overload:
        if (osig := _sig.get(overload)) is not None:  # type: ignore
            sig = CFUNCTYPE(osig.restype, *osig.argtypes)
        else:
            # Need to fallback on something. Raise a warning that no
            # overload was defined and that it will fallback to the
            # first entry in the dict.
            first = list(_sig.items())[0]
            calling_logger.warning(f"No function arguments overload was provided for {name}. ")
            calling_logger.warning(f"Falling back to the first overload ({first[0]})")
            sig = CFUNCTYPE(first[1].restype, *first[1].argtypes)
    else:
        raise ValueError(
            f"Invalid data type provided for `sig` argument: {type(_sig)}. Must be one of FUNCDEF or a "
            "dictionary containing a mapping of FUNCDEF objects representing overloads."
        )
    if isinstance(offset, dict):
        # Handle overloads
        if (_offset := offset.get(overload)) is not None:  # type: ignore
            offset = _offset
        else:
            _offset = list(offset.items())[0]
            calling_logger.warning(f"No function arguments overload was provided for {name}. ")
            calling_logger.warning(f"Falling back to the first overload ({_offset[0]})")
            offset = _offset[1]
    binary_base = _internal.BASE_ADDRESS
    # TODO: This is inefficient to look it up every time. This should be optimised at some point.
    if binary is not None:
        if (hm := _get_binary_info(binary)) is not None:
            _, module = hm
            binary_base = module.lpBaseOfDll

    cfunc = sig(binary_base + offset)
    return cfunc(*args)



# /// script
# dependencies = ["pymhf[gui]>=0.1.16"]
#
# [tool.pymhf]
# exe = "../programs/app.exe"
# start_exe = true
# start_paused = true
# interactive_console = false
# args = [2, 3]
#
# [tool.pymhf.gui]
# always_on_top = true
#
# [tool.pymhf.logging]
# log_dir = "."
# log_level = "info"
# window_name_override = "Test mod"
# ///
from __future__ import annotations

import ctypes
import logging

from pymhf import Mod
from pymhf.core.hooking import static_function_hook

logger = logging.getLogger("TestLogger")


@static_function_hook(exported_name="multiply")
def multiply(a: ctypes.c_int32, b: ctypes.c_int64) -> ctypes.c_int64: ...


class AppMod(Mod):
    @multiply.before
    def before_multiply(self, a: int, b: int):
        # Let's double the inputs
        logger.info(f"Doubling the inputs {a} and {b}")
        return 2 * a, 2 * b

    @multiply.after
    def after_multiply(self, a: int, b: int, _result_: int):
        logger.info(f"({a} * 2) * ({b} * 2) = {_result_}")



import os.path as op
from subprocess import Popen

from pymhf import run_module

if __name__ == "__main__":
    # Run a mod by creating the process and then attaching pymhf later.
    # It is generally recommended that you just use pyMHF's built-in functionality for doing this.
    proc = Popen("notepad.exe")
    CONFIG = {
        "pid": proc.pid,
        "start_paused": False,
        "start_exe": False,
    }
    run_module(op.join(op.dirname(__file__), "test_mod.py"), CONFIG, None, None)



import asyncio
import concurrent.futures
import glob
import os
import os.path as op
import subprocess
import time
import webbrowser
from functools import partial
from signal import SIGTERM
from threading import Event
from typing import Optional

import psutil
import pymem
import pymem.exception
import pymem.process
import pymem.ressources.kernel32
import pyrun_injected.dllinject as dllinject

from pymhf.core._types import LoadTypeEnum, pymhfConfig
from pymhf.core.hashing import hash_bytes_from_file, hash_bytes_from_memory
from pymhf.core.importing import parse_file_for_mod
from pymhf.core.log_handling import open_log_console
from pymhf.core.process import start_process
from pymhf.core.protocols import ESCAPE_SEQUENCE, TerminalProtocol
from pymhf.utils.config import canonicalize_setting
from pymhf.utils.parse_toml import read_pymhf_settings
from pymhf.utils.winapi import get_exe_path_from_pid

CWD = op.dirname(__file__)
PYMHF_DIR = op.dirname(CWD)
APPDATA_DIR = os.environ.get("APPDATA", op.expanduser("~"))

# Flag indicating whether to kill our own pid.
# This should generally be true except when running tests otherwise it will kill the test runner.
REMOVE_SELF = True
END_EVENT = Event()


class WrappedProcess:
    def __init__(self, proc: Optional[psutil.Process] = None, thread_handle: Optional[int] = None):
        self.proc = proc
        self.thread_handle = thread_handle
        self._is_self_started = False
        if self.thread_handle is not None:
            self._is_self_started = True

    def suspend(self):
        if self.proc is not None:
            self.proc.suspend()

    def resume(self):
        if self.proc is not None:
            self.proc.resume()
        else:
            pymem.ressources.kernel32.ResumeThread(self.thread_handle)  # type: ignore


class pymhfExitException(Exception):
    pass


def _wait_until_process_running(target: str):
    run = True
    while run:
        try:
            for p in psutil.process_iter(["name", "pid"]):
                if p.name().lower() == target.lower():
                    return WrappedProcess(proc=p)
        except KeyboardInterrupt:
            return None


def get_process_when_ready(
    cmd: list[str],
    target: str,
    required_assemblies: Optional[list[str]] = None,
    is_steam: bool = True,
    start_paused: bool = False,
) -> tuple[Optional[dllinject.pyRunner], Optional[WrappedProcess]]:
    target_process: Optional[WrappedProcess] = None
    parent_process = None
    # If we are running something which is under steam, make sure steam is
    # running first.
    if is_steam:
        for p in psutil.process_iter(["name", "pid"]):
            if p.name().lower() == "steam.exe":
                parent_process = p
                break
        if parent_process is None:
            raise ProcessLookupError("Steam not running! For now, start it yourself and try again...")

    run = True
    found_pid = None
    if parent_process is not None:
        if is_steam:
            # For steam games, if we have the game ID then the cmd will be
            # steam://rungameid/{game id} which we need to invoke this way.
            print("Running from steam")
            # We can only run the first argument in the list with steam unfortunately...
            webbrowser.open(cmd[0])
        else:
            subprocess.run(cmd)
        while run:
            try:
                if target_process is None:
                    for child in parent_process.children(recursive=True):
                        if child.name() == target:
                            target_process = WrappedProcess(proc=child)
                else:
                    binary = pymem.Pymem(target, exact_match=True)
                    modules = list(pymem.process.enum_process_module(binary.process_handle))
                    if len(modules) > 0 and required_assemblies is not None:
                        if set(required_assemblies) <= set(x.name for x in modules):
                            run = False
                            break
            except KeyboardInterrupt:
                raise
            except psutil.NoSuchProcess:
                # Race-condition case where steam creates some short-lived process which dies before psutil
                # can handle it properly.
                pass
    else:
        creationflags = 0x4 if start_paused else 0
        process_handle, thread_handle, found_pid, tid = start_process(  # noqa
            cmd, creationflags=creationflags
        )
        target_process = WrappedProcess(thread_handle=thread_handle)

    if target_process is not None:
        if found_pid is not None:
            binary = pymem.Pymem(found_pid)
        else:
            binary = pymem.Pymem(target, exact_match=True)
        injected = dllinject.pyRunner(binary)
        if start_paused:
            target_process.suspend()

        return injected, target_process
    else:
        return None, None


def load_mod_file(filepath: str, config_overrides: Optional[pymhfConfig] = None):
    """Load an individual file as a mod.

    Parameters
    ----------
    filepath
        The relative or absolute filepath to a single python file to be loaded as a mod.
        It is generally recommended to pass an absolute path as it will be more reliable.
    config_overrides
        An optional dictionary containing values which will override any existing config values read from the
        specified file.
    """
    pymhf_settings = read_pymhf_settings(filepath, True)
    if config_overrides:
        pymhf_settings.update(config_overrides)
    run_module(filepath, pymhf_settings, None, None)


def load_module(plugin_name: str, module_path: str):
    """Load a module.
    This should be used when loading an entire folder or a plugin based on name.
    """
    config_dir = op.join(APPDATA_DIR, "pymhf", plugin_name)
    local_cfg_file = op.join(config_dir, "pymhf.local.toml")
    if op.exists(local_cfg_file):
        local_cfg = read_pymhf_settings(local_cfg_file).get("local_config", {})
    else:
        local_cfg = {}
    module_cfg_file = op.join(module_path, "pymhf.toml")
    module_cfg = read_pymhf_settings(module_cfg_file)
    module_cfg.update(local_cfg)
    run_module(module_path, module_cfg, plugin_name, config_dir)


def run_module(
    module_path: str,
    config: pymhfConfig,
    plugin_name: Optional[str] = None,
    config_dir: Optional[str] = None,
):
    """Run the module provided.

    Parameters
    ----------
    module_path
        The path to the module or single-file mod to be run.
    config
        A mapping of the associated pymhf config.
    plugin_name
        The name of the plugin. This will only be provided if we are running a library.
    config_dir
        The local config directory. This will only be provided if we are running a library.
    """
    if plugin_name is None:
        if op.isfile(module_path):
            load_type = LoadTypeEnum.SINGLE_FILE
        if op.exists(op.join(module_path, "pymhf.toml")):
            load_type = LoadTypeEnum.MOD_FOLDER
    else:
        load_type = LoadTypeEnum.LIBRARY

    binary_path = None
    injected = None
    binary_exe = config.get("exe", None)
    required_assemblies = []
    cmd: list[str] = []
    required_assemblies = config.get("required_assemblies", [])
    start_exe = config.get("start_exe", True)
    cmd_args = config.get("args", [])
    if not isinstance(cmd_args, list):
        print(f"Warning: The provided args value {cmd_args!r} is not valid. It must be a list.")
        cmd_args = []
    # Ensure each value is a string.
    for i, _arg in enumerate(cmd_args):
        cmd_args[i] = str(_arg)
    to_load_pid: Optional[int] = config.get("pid", None)
    interactive_console = config.get("interactive_console", True)
    logging_config = config.get("logging", {}) or {}
    show_log_window = logging_config.get("shown", True)
    if to_load_pid is None and binary_exe is None:
        raise ValueError("[tool.pymhf] requires either an `exe` or `pid` value.")
    if binary_exe is None and start_exe is True:
        raise ValueError("[tool.pymhf] requires `exe` to be set if `start_exe = true`")
    try:
        steam_gameid = int(config.get("steam_gameid", 0))
    except (ValueError, TypeError):
        steam_gameid = 0
    start_paused = config.get("start_paused", False)

    if config_dir is None:
        cache_dir = op.join(module_path, ".cache")
    else:
        cache_dir = op.join(config_dir, ".cache")

    # Check if the module_path is a file or a folder.
    _module_path = module_path
    if op.isfile(module_path):
        _module_path = op.dirname(module_path)

    is_steam = False
    if steam_gameid:
        cmd = [f"steam://rungameid/{steam_gameid}"]
        is_steam = True
    elif binary_exe is not None:
        if op.isabs(binary_exe):
            binary_path = binary_exe
            cmd = [binary_path]
            # We only need the binary_exe to be the name from here on.
            binary_exe = op.basename(binary_exe)
        else:
            # Try and create the path relative to the provided module path.
            _trial_path = op.realpath(op.join(_module_path, binary_exe))
            if op.exists(_trial_path):
                binary_path = _trial_path
                cmd = [binary_path]
                binary_exe = op.basename(binary_path)
            else:
                if start_exe is True:
                    raise ValueError(
                        "[tool.pymhf].exe must be a full path or path relative to the running script if no "
                        "steam_gameid is provided and start_exe is not true"
                    )
                else:
                    pass

    if start_exe and binary_exe is not None:
        # Check to see fi the binary is already running... Just in case.
        # If it is, we use it, otherwise start it.
        try:
            pm_binary = pymem.Pymem(binary_exe, exact_match=True)
            print(f"Found an already running instance of {binary_exe!r}")
            injected = dllinject.pyRunner(pm_binary)
            print("Python injected")
            proc = None
        except (pymem.exception.ProcessNotFound, pymem.exception.CouldNotOpenProcess):
            injected, proc = get_process_when_ready(
                cmd + cmd_args,
                binary_exe,
                required_assemblies,
                is_steam,
                start_paused,
            )
    else:
        # If we aren't starting the exe then by the time we have run we expect the process to already exist,
        # so just find it with pymem.
        if binary_exe is not None:
            pm_binary = pymem.Pymem(binary_exe, exact_match=True)
            injected = dllinject.pyRunner(pm_binary)
            print("Python injected")
            proc = None
        else:
            pm_binary = pymem.Pymem(to_load_pid)  # type: ignore  (pymem has the wrong type...)
            injected = dllinject.pyRunner(pm_binary)
            print("Python injected")
            proc = None

    if injected is None:
        # TODO: Raise better error messages/reason why it couldn't load.
        print("FATAL ERROR: Cannot start process!")
        return
    if injected.pm is None:
        print("FATAL ERROR: Cannot start process!")
        return

    pm_binary = injected.pm
    if pm_binary.process_id is None:
        print("FATAL ERROR: Cannot start process!")
        return

    if binary_path is None:
        binary_path = get_exe_path_from_pid(pm_binary)

    binary_dir = None
    if binary_path is not None:
        binary_dir = op.dirname(binary_path)

    print(f"Found PID: {pm_binary.process_id}")

    log_pid = None
    logging_config = config.get("logging", {})
    log_window_name_override = logging_config.get("window_name_override", "pymhf console")
    _log_dir = logging_config.get("log_dir", "{CURR_DIR}")
    log_dir = canonicalize_setting(_log_dir, plugin_name, _module_path, binary_dir, "logs")
    if log_dir is None and binary_dir is not None:
        log_dir = op.join(binary_dir, "LOGS")
    if log_dir is None:
        print(f"Unable to determine a location to put logs from {_log_dir}. No logs will be saved!")

    mod_save_dir = config.get("mod_save_dir", "{CURR_DIR}")
    mod_save_dir = canonicalize_setting(mod_save_dir, plugin_name, _module_path, binary_dir, "MOD_SAVES")

    executor = None
    futures = []
    try:
        loop = asyncio.get_event_loop()
    except (RuntimeError, ValueError):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    def kill_injected_code(loop: asyncio.AbstractEventLoop):
        # End one last "escape sequence" message:
        client_completed = asyncio.Future()
        client_factory = partial(TerminalProtocol, message=ESCAPE_SEQUENCE.decode(), future=client_completed)
        factory_coroutine = loop.create_connection(
            client_factory,
            "127.0.0.1",
            6770,
        )
        loop.run_until_complete(factory_coroutine)
        loop.run_until_complete(client_completed)

    try:
        if log_dir and show_log_window:
            log_pid = open_log_console(op.join(CWD, "log_terminal.py"), log_dir, log_window_name_override)
        # Have a small nap just to give it some time.
        time.sleep(0.5)
        if binary_path:
            try:
                with open(binary_path, "rb") as f:
                    binary_hash = hash_bytes_from_file(f)
            except PermissionError:
                print(f"Cannot open {binary_path!r} to hash it. Trying to read from memory...")
                binary_hash = hash_bytes_from_memory(pm_binary)
            print(f"Exe hash is: {binary_hash}")
        else:
            binary_hash = 0

        def close_callback(x):
            print("pyMHF exiting...")
            for _pid in {pm_binary.process_id, log_pid}:
                if _pid:
                    try:
                        os.kill(_pid, SIGTERM)
                    except Exception:
                        # If we can't kill it, it's probably already dead. Just continue.
                        pass
            END_EVENT.set()
            # Finally, send a SIGTERM to ourselves...
            if REMOVE_SELF:
                os.kill(os.getpid(), SIGTERM)

        # Wait some time for the data to be written to memory only if we are starting the process ourselves.
        if start_exe:
            time.sleep(2)

        offset_map = {}
        # This is a mapping of the required assemblies to their filenames so we may do an import look up on
        # the inside.
        included_assemblies = {}
        modules = []
        if required_assemblies:
            modules = list(pm_binary.list_modules())
            found_modules = list(filter(lambda x: x.name in required_assemblies, modules))
            if not found_modules:
                print(f"Cannot find specified assembly from config ({required_assemblies})")
                return
            for module in found_modules:
                offset_map[module.name] = (module.lpBaseOfDll, module.SizeOfImage)
                included_assemblies[module.name] = module.filename
        else:
            pb = pm_binary.process_base
            offset_map[binary_exe] = (pb.lpBaseOfDll, pb.SizeOfImage)
        # if not modules:
        #     modules = pm_binary.list_modules()
        # for module in modules:
        #     included_assemblies[module.name] = module.filename
        # print(f"The handle: {pm_binary.process_handle}, bases: {offset_map}")

        if len(offset_map) == 1:
            # Only one thing. For now, we just set these as the `binary_base` and `binary_size`.
            # TODO: When we want to support offsets and hooks in multiple assemblies we need to pass the whole
            # dictionary in potentially, or do a lookup from inside the process.
            assem_name = list(offset_map.keys())[0]
            binary_base = offset_map[assem_name][0]
            binary_size = offset_map[assem_name][1]

        try:
            cwd = CWD.replace("\\", "\\\\")
            import sys

            _path = sys.path
            _path.insert(0, PYMHF_DIR)

            # Inject the folder the mod is in (or the folder being run) if it's not already in the sys.path
            if load_type == LoadTypeEnum.SINGLE_FILE:
                _mod_dir = op.dirname(module_path)
                if _mod_dir not in _path:
                    _path.insert(0, _mod_dir)
            if load_type == LoadTypeEnum.MOD_FOLDER:
                if module_path not in _path:
                    _path.insert(0, module_path)
            if load_type == LoadTypeEnum.LIBRARY:
                # For libraries, the module_path will be the directory where the pymhf.toml file is located
                # which is one path deeper than we want to inject into the path.
                if (mp := op.dirname(module_path)) not in _path:
                    _path.insert(0, mp)
                # Also add the mod directory to the path.
                _mod_folder = config.get("mod_dir")
                mod_folder = canonicalize_setting(_mod_folder, "pymhf", _module_path, binary_dir)
                if _mod_folder and not mod_folder:
                    # In this case the directory for the mod folder doesn't actually exist. Log the steps
                    # required to re-configure.
                    print(
                        f"[ERROR] Mod folder can't be found or resolved: {_mod_folder}.\n"
                        "[ERROR] Please reconfigure the path by running `pymhf config <library or path>` and "
                        "then selecting 'Set mod directory'."
                    )
                    sys.exit(0)
                if mod_folder:
                    if mod_folder not in _path:
                        _path.insert(0, mod_folder)
                    # Loop over each of the direct descendents of the mod folder and see if each folder
                    # contains any mods.
                    # If it does, then inject that directory into the path.
                    for _fpath in os.listdir(mod_folder):
                        fpath = op.join(mod_folder, _fpath)
                        if op.isdir(fpath):
                            for pyfile in glob.iglob(op.join(fpath, "*.py")):
                                with open(pyfile, "r") as f:
                                    if parse_file_for_mod(f.read()):
                                        # Once we know that at least one file contains a mod, stop iterating
                                        # and add the folder to the path.
                                        if fpath not in _path:
                                            _path.insert(0, fpath)
                                        break
            saved_path = [x.replace("\\", "\\\\") for x in _path]

            injected_data_list = []

            # Inject the new path
            sys_path_str = f"""
import sys
sys.path = {saved_path}
"""
            injected_data_list.append(dllinject.StringType(sys_path_str, False))

            # Inject our preinject script.
            injected_data_list.append(dllinject.StringType(op.join(CWD, "_preinject.py"), True))

            # Allocate a boolean value which will be set to True by the injected code once it has completed.
            sentinel_addr = pm_binary.allocate(4)
            pm_binary.write_bool(sentinel_addr, False)

            # Inject the common NMS variables which are required for general use.
            internals_str = f"""
pymhf.core._internal.MODULE_PATH = {module_path!r}
pymhf.core._internal.BASE_ADDRESS = {binary_base!r}
pymhf.core._internal.SIZE_OF_IMAGE = {binary_size!r}
pymhf.core._internal.CWD = {cwd!r}
pymhf.core._internal.PID = {pm_binary.process_id!r}
pymhf.core._internal.HANDLE = {pm_binary.process_handle!r}
pymhf.core._internal.BINARY_HASH = {binary_hash!r}
pymhf.core._internal.CONFIG = {config!r}
pymhf.core._internal.EXE_NAME = {binary_exe!r}
pymhf.core._internal.BINARY_PATH = {binary_path!r}
pymhf.core._internal.LOAD_TYPE = {load_type.value!r}
pymhf.core._internal.MOD_SAVE_DIR = {mod_save_dir!r}
pymhf.core._internal.INCLUDED_ASSEMBLIES = {included_assemblies!r}
pymhf.core._internal.CACHE_DIR = {cache_dir!r}
pymhf.core._internal._SENTINEL_PTR = {sentinel_addr!r}
                """
            injected_data_list.append(dllinject.StringType(internals_str, False))
        except Exception as e:
            import traceback

            print(e)
            print(traceback.format_exc())
        # Inject the script
        injected_data_list.append(dllinject.StringType(op.join(CWD, "injected.py"), True))
        executor = concurrent.futures.ThreadPoolExecutor(max_workers=2)
        fut = executor.submit(injected.run_data, injected_data_list)
        fut.add_done_callback(lambda x: close_callback(x))
        futures.append(fut)

        # Wait for the injected process to indicate that it's ready to go.
        if start_paused and start_exe is not False:
            while True:
                try:
                    if pm_binary.read_bool(sentinel_addr) is True:
                        print("pyMHF injection complete!")
                        break
                except KeyboardInterrupt:
                    # Kill the injected code even though we are still waiting for it to start up.
                    kill_injected_code(loop)
                    raise
                except pymem.exception.MemoryReadError:
                    # In this case the process has probably already died for some reason...
                    # Set END_EVENT so that we just initiate the shutdown process.
                    END_EVENT.set()
            if proc is not None:
                proc.resume()
                print("Press CTRL+C to exit the process:\n")

        if interactive_console:
            print("pyMHF interactive python command prompt")
            print("Type any valid python commands to execute them within the games' process")
            # TODO: change this to use a threading.Event object, and in a separate thread, poll every second
            # or so the running game process to see if it's still running (using psutil?)
            # Might need to modify the WrappedProcess object to always create the pstuil.Process object to
            # make it work easier...

            # TODO: This should become False when we exit from the program...
            while True:
                try:
                    input_ = input(">>> ")
                    client_completed = asyncio.Future()
                    client_factory = partial(TerminalProtocol, message=input_, future=client_completed)
                    factory_coroutine = loop.create_connection(
                        client_factory,
                        "127.0.0.1",
                        6770,
                    )
                    loop.run_until_complete(factory_coroutine)
                    loop.run_until_complete(client_completed)
                except KeyboardInterrupt:
                    break
            kill_injected_code(loop)
        else:
            while True:
                # Wait until the end event is triggered. As soon as it is this loop will exit.
                try:
                    if END_EVENT.wait(2) is True:
                        break
                except KeyboardInterrupt:
                    break
            kill_injected_code(loop)
    except KeyboardInterrupt:
        # If it's a keyboard interrupt, just pass as it will have bubbled up from
        # below.
        pass
    except Exception as e:
        # Any other exception we want to actually know about.
        import traceback

        print(e)
        print(traceback.format_exc())
        raise
    finally:
        loop.close()
        try:
            for _ in concurrent.futures.as_completed(futures, timeout=5):
                pass
        except TimeoutError:
            # Don't really care.
            print("Got a time out error...")
            pass
        if executor is not None:
            executor.shutdown(wait=False)

        # Have a short nap and then finish trying to clean up.
        time.sleep(0.5)
        if REMOVE_SELF:
            pids = {pm_binary.process_id, log_pid}
        else:
            pids = {
                log_pid,
            }
        for _pid in pids:
            if _pid:
                try:
                    os.kill(_pid, SIGTERM)
                    print(f"Just killed process {_pid}")
                except Exception:
                    # If we can't kill it, it's probably already dead. Just continue.
                    print(f"Failed to kill process {_pid}. It was likely already dead...")
                    pass



import ctypes
import inspect
from dataclasses import dataclass
from typing import Optional, Type, TypeVar, Union, _AnnotatedAlias, get_args

from typing_extensions import get_type_hints

from pymhf.extensions.ctypes import c_enum32

_T = TypeVar("_T", bound=Type[ctypes.Structure])

CTYPES = Union[
    ctypes._SimpleCData,
    ctypes.Structure,
    ctypes._Pointer,
    ctypes._Pointer_orig,  # The original, un-monkeypatched ctypes._Pointer object
    ctypes.Array,
    ctypes.Union,
    c_enum32,
]


@dataclass
class Field:
    datatype: CTYPES
    offset: Optional[int] = None


def partial_struct(cls: _T) -> _T:
    """Mark the decorated class as a partial struct.
    This will automatically construct the ``_field_`` attribute for this class
    """
    # Always get the calling frame and add the locals so that if we have any annotated fields they won't fail.
    # In the case of a struct which has no annotations this will be ever so slightly slower than if we did a
    # try... except.
    # However in the case of annotated fields, it's much faster, so it's beneficial to always do this.
    calling_frame = None
    _locals = locals()
    if (cframe := inspect.currentframe()) is not None:
        calling_frame = cframe.f_back
    if calling_frame is not None:
        _locals.update(calling_frame.f_locals)

    # Also add the class to the locals so that it can find itself in the case it references itself.
    _locals.update({cls.__name__: cls})

    _fields_ = []
    curr_position = 0
    total_size = getattr(cls, "_total_size_", 0)
    # If there are no annotations, it's just an empty ctypes.Structure class.
    if not hasattr(cls, "__annotations__"):
        cls._fields_ = _fields_
        return cls
    # List of field names to exclude.
    # These are to ensure that when a partial struct subclasses from another it doesn't get the fields twice.
    exclude_fields = set()
    subclass_size = 0
    for subclass in cls.__mro__[1:]:
        if hasattr(subclass, "_fields_"):
            for field in subclass._fields_:
                exclude_fields.add(field[0])
            subclass_size += ctypes.sizeof(subclass)
    # If there are, loop over the annotations and extract the info we need to construct the _fields_.
    for field_name, annotation in get_type_hints(cls, include_extras=True, localns=_locals).items():
        # Ingore any fields which we have picked up from any subclasses.
        if field_name in exclude_fields:
            continue
        if not isinstance(annotation, _AnnotatedAlias):
            # In this case it's just the type.
            field_type = annotation
            field_offset = None
        else:
            # If the annotation is a Field object, get info from it, otherwise it must be an integer
            # specifying the offset.
            metadata = annotation.__metadata__
            if len(metadata) != 1:
                raise ValueError(f"The field {field_name!r} has an invalid annotation: {annotation}")
            metadata = metadata[0]
            if isinstance(metadata, Field):
                field_type = metadata.datatype
                field_offset = metadata.offset
            else:
                field_type = annotation.__origin__
                field_offset = metadata
        # Check to make sure that the `field_type` is not a string(as can happen when we have a
        # self-reference).
        if isinstance(field_type, str):
            field_type = getattr(annotation, "__origin__", None)
            if field_type is None:
                raise ValueError(
                    f"The provided metadata {metadata} for field {field_name!r} is invalid. "
                    "If the type in the `Field` component of the annotation is a string, please ensure the "
                    "first argument of the Annotation is also the same string so that the type can be "
                    "resolved."
                )
        if not issubclass(field_type, get_args(CTYPES)):
            raise ValueError(f"The field {field_name!r} has an invalid type: {field_type}")
        if field_offset is not None and not isinstance(field_offset, int):
            raise ValueError(f"The field {field_name!r} has an invalid offset: {field_offset!r}")
        # Correct the field offset by the size of the subclass.
        if field_offset and field_offset - subclass_size > curr_position:
            padding_bytes = field_offset - subclass_size - curr_position
            _fields_.append((f"_padding_{curr_position:X}", ctypes.c_ubyte * padding_bytes))
            curr_position += padding_bytes
        field_alignment = ctypes.alignment(field_type)
        if field_alignment and (curr_position % field_alignment != 0):
            # If the field is not aligned to the correct position, then move the current position forward
            # to ensure it's right.
            # Don't bother adding this as padding since it will just add extra unnecessary fields.
            diff = field_alignment - (curr_position % field_alignment)
            curr_position += diff
        _fields_.append((field_name, field_type))
        curr_position += ctypes.sizeof(field_type)
    if total_size and curr_position < total_size:
        padding_bytes = total_size - curr_position
        _fields_.append((f"_padding_{curr_position:X}", ctypes.c_ubyte * padding_bytes))
    cls._fields_ = _fields_
    return cls



import os
import os.path as op
import tempfile

import pytest

import pymhf

pytestmark = pytest.mark.skipif(os.environ.get("CI") is not None, reason="Flaky on CI")

MODS_DIR = op.realpath(op.join(op.dirname(__file__), "..", "test_mods"))
APPS_DIR = op.realpath(op.join(op.dirname(__file__), "..", "programs"))


def test_default_args():
    with tempfile.TemporaryDirectory() as tempdir:
        config = {
            "exe": op.join(APPS_DIR, "app.exe"),
            "start_paused": True,
            "start_exe": True,
            "interactive_console": False,
            "logging": {"log_dir": tempdir},
        }
        pymhf.main.REMOVE_SELF = False
        pymhf.load_mod_file(op.join(MODS_DIR, "app_test_mod.py"), config)

        log_files = os.listdir(tempdir)
        assert len(log_files) == 1
        log_file = log_files[0]
        important_lines = []
        with open(op.join(tempdir, log_file), "r") as f:
            for line in f:
                if "TestLogger" in line:
                    idx = line.find("INFO")
                    actual_val = line[idx + 4 :]
                    important_lines.append(actual_val.strip())
        assert important_lines == ["Doubling the inputs 2 and 3", "(2 * 2) * (3 * 2) = 24"]



import logging
from enum import Enum
from typing import Optional, TypedDict, Union

# import win32con
import dearpygui.dearpygui as dpg
import win32gui

import pymhf.core._internal as _internal
import pymhf.core.caching as cache
from pymhf.core._types import pymhfConfig
from pymhf.core.mod_loader import Mod, ModManager
from pymhf.gui.hexview import HexView
from pymhf.gui.protocols import ButtonProtocol, ComboBoxProtocol, VariableProtocol, VariableType
from pymhf.utils.winapi import set_window_transparency

SETTINGS_NAME = "_pymhf_gui_settings"
HEX_NAME = "_pymhf_gui_hex"
DETAILS_NAME = "_pymhf_gui_details"
WINDOW_TITLE = "pyMHF"

# TODO:
# - add keyboard shortcut to show or hide the GUI
# - Add support for mod states


logger = logging.getLogger(__name__)
rootLogger = logging.getLogger("")


class WidgetType(Enum):
    BUTTON = 0
    VARIABLE = 1
    TEXT = 2
    COMBOBOX = 3


class Widgets(TypedDict):
    buttons: dict[str, Union[int, str]]
    comboboxes: dict[str, list[tuple[Union[int, str], WidgetType]]]
    variables: dict[str, list[tuple[Union[int, str], WidgetType]]]


def toggle_on_top(item: int, value: bool):
    dpg.set_viewport_always_top(value)


class GUI:
    hex_view: HexView

    def __init__(self, mod_manager: ModManager, config: pymhfConfig):
        self.config = config
        self.always_on_top = config.get("gui", {}).get("always_on_top", False)
        self.is_debug = config.get("logging", {}).get("log_level") == "debug"
        self.scale = config.get("gui", {}).get("scale", 1)
        dpg.create_context()
        dpg.create_viewport(
            title=WINDOW_TITLE,
            width=int(600 * self.scale),
            height=int(800 * self.scale),
            decorated=True,
            always_on_top=self.always_on_top,
        )
        dpg.setup_dearpygui()
        dpg.set_global_font_scale(self.scale)
        self.tracking_variables: dict[str, list[tuple[str, Mod, str, bool]]] = {}
        self.tabs: dict[int, str] = {}
        self._shown = True
        self._handle = None
        self._current_tab: Optional[str] = None
        self.mod_manager = mod_manager

        self.widgets: dict[str, Widgets] = {}

        # Keep track of the viewport dimensions and position.
        # NOTE: These are ONLY updated when the viewport is minimised by the `hide_window` method.
        # TODO: Maybe set a max height and width?
        self._window_dimensions = [0, 0]
        self._window_position: list[float] = [0, 0]

        # Some info related settings
        self._hide_pyd_modules = True

        self.add_window()

    def alpha_callback(self, sender, app_data):
        set_window_transparency(self.hwnd, app_data)

    def show_window(self):
        # TODO: This needs to be called twice to run properly...
        # It might still be better to try and use the `win32gui` calls to get the system to handle this better
        if self._window_dimensions != [0, 0]:
            dpg.maximize_viewport()
            dpg.set_viewport_width(self._window_dimensions[0])
            dpg.set_viewport_height(self._window_dimensions[1])
            dpg.set_viewport_pos(self._window_position)
        # if not self._shown and self._handle:
        #     win32gui.ShowWindow(self._handle, win32con.SW_SHOW)
        #     win32gui.SetWindowLong(self._handle, win32con.GWL_EXSTYLE, 0)
        #     self._shown = True

    def hide_window(self):
        self._window_dimensions = (dpg.get_viewport_width(), dpg.get_viewport_height())
        self._window_position = dpg.get_viewport_pos()
        dpg.minimize_viewport()
        # if self._shown and self._handle:
        #     win32gui.ShowWindow(self._handle, win32con.SW_HIDE)
        #     self._shown = False

    def toggle_debug_mode(self, _sender, is_debug):
        try:
            self.config["logging"]["log_level"] = is_debug and "debug" or "info"
        except KeyError:
            pass
        if is_debug:
            rootLogger.setLevel(logging.DEBUG)
        else:
            rootLogger.setLevel(logging.INFO)

    def toggle_show_gui(self, _sender, show_gui):
        self.config["gui"]["shown"] = show_gui

    def add_hex_tab(self):
        tab = dpg.add_tab(label="Hex View", tag=HEX_NAME, parent="tabbar")
        tab_alias = dpg.get_alias_id(tab)
        self.tabs[tab_alias] = HEX_NAME

        self.hex_view = HexView(HEX_NAME)
        self.hex_view._setup()

    def add_settings_tab(self):
        """Add a settings tab to configure the gui and other things."""
        with dpg.value_registry():
            dpg.add_bool_value(tag="always_on_top", default_value=self.always_on_top)
            dpg.add_bool_value(tag="is_debug", default_value=self.is_debug)
            dpg.add_bool_value(tag="show_gui", default_value=True)
        tab = dpg.add_tab(label="Settings", tag=SETTINGS_NAME, parent="tabbar")
        tab_alias = dpg.get_alias_id(tab)
        self.tabs[tab_alias] = SETTINGS_NAME

        with dpg.table(header_row=False, parent=SETTINGS_NAME, policy=dpg.mvTable_SizingStretchProp):
            dpg.add_table_column()
            dpg.add_table_column()

            # Toggle for debug mode
            with dpg.table_row():
                dpg.add_text("Enable debug mode")
                dpg.add_checkbox(
                    source="is_debug",
                    callback=self.toggle_debug_mode,
                )

            # Toggle for whether to show the gui at all.
            with dpg.table_row():
                dpg.add_text("Show GUI")
                dpg.add_checkbox(
                    source="show_gui",
                    callback=self.toggle_show_gui,
                )

            # Add a slider for the visibility.
            with dpg.table_row():
                dpg.add_text("Transparency")
                dpg.add_slider_float(
                    default_value=1,
                    max_value=1,
                    min_value=0,
                    callback=self.alpha_callback,
                )

            # Add a checkbox to toggle whether the window should stay always on top.
            with dpg.table_row():
                dpg.add_text("Always on top")
                dpg.add_checkbox(
                    source="always_on_top",
                    callback=toggle_on_top,
                )

    def _toggle_show_pyd(self, item: int, value: bool):
        self._hide_pyd_modules = value

    def add_details_tab(self):
        tab = dpg.add_tab(label="Details", tag=DETAILS_NAME, parent="tabbar")
        tab_alias = dpg.get_alias_id(tab)
        self.tabs[tab_alias] = DETAILS_NAME

        imports = _internal.imports
        tree = dpg.add_tree_node(label="Imports", parent=DETAILS_NAME)
        for dll_name, functions in imports.items():
            dll_branch = dpg.add_tree_node(label=dll_name, parent=tree)
            for func_name in functions.keys():
                dpg.add_tree_node(label=func_name, parent=dll_branch, leaf=True, bullet=True)

        # TODO: Add this toggle back and make it work.
        # with dpg.group(horizontal=True, parent=DETAILS_NAME):
        #     dpg.add_text("Hide *.pyd files")
        #     dpg.add_checkbox(callback=self._toggle_show_pyd, default_value=True)

        # Add in a section to show the actual list of loaded modules
        module_tree = dpg.add_tree_node(label="Loaded modules", parent=DETAILS_NAME)
        if self._hide_pyd_modules:
            names = (name for name in cache.module_map.keys() if not name.endswith(".pyd"))
        else:
            names = (name for name in cache.module_map.keys())
        for func_name in names:
            dpg.add_tree_node(label=func_name, parent=module_tree, leaf=True, bullet=True)

    def reload_tab(self, cls: Mod):
        """Reload the tab for the specific mod."""
        name = cls.__class__.__name__
        cls.pymhf_gui = self
        widgets = self.widgets.get(name, {})

        self.reload_buttons(cls, widgets)
        self.reload_comboboxes(cls, widgets)
        self.reload_variables(cls, widgets)

    def reload_buttons(self, cls: Mod, widgets: Widgets):
        """Reload all the buttons. Any new buttons will be added, any old ones will be removed, and any that
        remain will be reconfigured to point to the new bound method with any modified parameters (such as
        button label.)
        """
        button_widgets = widgets["buttons"]

        existing_button_names = set([b for b in button_widgets.keys()])
        mod_button_names = set([b for b in cls._gui_buttons.keys()])

        new_buttons = mod_button_names - existing_button_names
        for button_name in new_buttons:
            self.add_button(cls._gui_buttons[button_name])

        remaining_buttons = existing_button_names & mod_button_names
        for button_name in remaining_buttons:
            _button = cls._gui_buttons[button_name]
            dpg.configure_item(
                button_widgets[button_name],
                label=_button._button_text,
                callback=_button,
            )

        removed_buttons = existing_button_names - mod_button_names
        for button_name in removed_buttons:
            dpg.delete_item(button_widgets[button_name])
            button_widgets.pop(button_name)

    def reload_comboboxes(self, cls: Mod, widgets: Widgets):
        """Reload all the comboboxes. Any new combos will be added, any old ones will be removed, and any that
        remain will be reconfigured to point to the new bound method with any modified parameters (such as
        combo label and items).
        """
        combo_widgets = widgets["comboboxes"]

        existing_combo_names = set([c for c in combo_widgets.keys()])
        mod_combo_names = set([c for c in cls._gui_comboboxes.keys()])

        new_combos = mod_combo_names - existing_combo_names
        for combo_name in new_combos:
            self.add_combobox(cls._gui_comboboxes[combo_name])

        remaining_combos = existing_combo_names & mod_combo_names
        for combo_name in remaining_combos:
            _combo = cls._gui_comboboxes[combo_name]
            # Combo boxes have their text prepended.
            for var_id, var_type in combo_widgets[combo_name]:
                if var_type == WidgetType.TEXT:
                    dpg.set_value(var_id, _combo._combobox_text)
                else:
                    dpg.configure_item(var_id, items=_combo._items, callback=_combo)

        removed_combos = existing_combo_names - mod_combo_names
        for combo_name in removed_combos:
            for var_id, _ in combo_widgets[combo_name]:
                dpg.delete_item(var_id)
            combo_widgets.pop(combo_name)

    def reload_variables(self, cls: Mod, widgets: Widgets):
        """Reload all variables associated with a mod.
        Note that this will not work for changing an existing variables' type.
        To do this, remove (comment out) the property, reload the mod, and then uncomment, change the type and
        reload again to have it add it back.
        """
        variable_widgets = widgets["variables"]

        existing_variable_names = set([v for v in variable_widgets.keys()])
        mod_variable_names = set([v for v in cls._gui_variables.keys()])

        new_variables = mod_variable_names - existing_variable_names
        for variable_name in new_variables:
            self.add_variable(cls, variable_name, cls._gui_variables[variable_name])

        remaining_variables = existing_variable_names & mod_variable_names
        for variable_name in remaining_variables:
            # Loop over the text and variable ids so we may update them.
            for var_id, var_type in variable_widgets[variable_name]:
                if var_type == WidgetType.TEXT:
                    dpg.set_value(var_id, cls._gui_variables[variable_name]._label_text)
                else:
                    dpg.configure_item(var_id, user_data=(cls, variable_name))

        removed_variables = existing_variable_names - mod_variable_names
        for variable_name in removed_variables:
            # Remove the variable itself which dpg uses to store the value, as well as all widgets (text and
            # inputs) associated with it.
            name = cls.__class__.__name__
            tag = f"{name}.{variable_name}"
            dpg.delete_item(tag)
            for var_id, _ in variable_widgets[variable_name]:
                dpg.delete_item(var_id)
            variable_widgets.pop(variable_name)

    def add_tab(self, cls: Mod):
        """Add the mod as a new tab in the GUI."""
        # Check to see if the `no_gui` decorator has been applied to the class.
        # If so, don't add it now.
        if getattr(cls, "_no_gui", False) is True:
            return

        name = cls.__class__.__name__
        cls.pymhf_gui = self

        tab = dpg.add_tab(label=name, tag=name, parent="tabbar")
        tab_alias = dpg.get_alias_id(tab)
        dpg.set_item_user_data(name, cls)
        self.tabs[tab_alias] = name
        self.widgets[name] = {
            "buttons": {},
            "variables": {},
            "comboboxes": {},
        }

        dpg.add_button(
            label="Reload Mod",
            callback=self.mod_manager._gui_reload,
            user_data=(cls._mod_name, self),
            parent=name,
        )

        dpg.add_separator(parent=name)

        for _button in cls._gui_buttons.values():
            self.add_button(_button)

        for _combobox in cls._gui_comboboxes.values():
            self.add_combobox(_combobox)

        for variable_name, getter in cls._gui_variables.items():
            self.add_variable(cls, variable_name, getter)

    def change_tab(self, sender: str, app_data: int):
        self._current_tab = self.tabs[app_data]

    def add_window(self):
        with dpg.window(
            label="pyMHF",
            width=int(600 * self.scale),
            height=int(800 * self.scale),
            tag="pyMHF",
            on_close=self.exit,
        ):
            dpg.add_tab_bar(tag="tabbar", callback=self.change_tab)

    def add_button(self, callback: ButtonProtocol):
        name = callback.__self__.__class__.__name__
        meth_name = callback.__qualname__
        button = dpg.add_button(label=callback._button_text, callback=callback, parent=name)
        self.widgets[name]["buttons"][meth_name] = button

    def add_combobox(self, callback: ComboBoxProtocol):
        name = callback.__self__.__class__.__name__
        meth_name = callback.__qualname__

        with dpg.group(horizontal=True, parent=name):
            txt_id = dpg.add_text(callback._combobox_text)
            combobox = dpg.add_combo(
                items=callback._items,
                callback=callback,
            )
            self.widgets[name]["comboboxes"][meth_name] = [(txt_id, WidgetType.TEXT)]
            self.widgets[name]["comboboxes"][meth_name].append((combobox, WidgetType.COMBOBOX))

    def add_variable_gui_elements(self, cls: Mod, variable: str, getter: VariableProtocol):
        name = cls.__class__.__name__
        tag = f"{name}.{variable}"

        def on_update(_, app_data, user_data):
            setattr(user_data[0], user_data[1], app_data)

        with dpg.group(horizontal=True, parent=name):
            input_id = None
            txt_id = dpg.add_text(getter._label_text)
            self.widgets[name]["variables"][variable] = [(txt_id, WidgetType.TEXT)]
            if getter._has_setter:
                if getter._variable_type == VariableType.INTEGER:
                    extra_args = {"on_enter": False}
                    extra_args.update(getter._extra_args)
                    input_id = dpg.add_input_int(
                        source=tag,
                        callback=on_update,
                        user_data=(cls, variable),
                        **extra_args,
                    )
                elif getter._variable_type == VariableType.STRING:
                    extra_args = {"on_enter": False}
                    extra_args.update(getter._extra_args)
                    input_id = dpg.add_input_text(
                        source=tag,
                        callback=on_update,
                        user_data=(cls, variable),
                        **extra_args,
                    )
                elif getter._variable_type == VariableType.FLOAT:
                    extra_args = {"on_enter": False}
                    extra_args.update(getter._extra_args)
                    input_id = dpg.add_input_double(
                        source=tag,
                        callback=on_update,
                        user_data=(cls, variable),
                        **extra_args,
                    )
                elif getter._variable_type == VariableType.BOOLEAN:
                    extra_args = {}
                    extra_args.update(getter._extra_args)
                    input_id = dpg.add_checkbox(
                        source=tag,
                        callback=on_update,
                        user_data=(cls, variable),
                        **extra_args,
                    )
                if input_id is not None:
                    self.widgets[name]["variables"][variable].append((input_id, WidgetType.VARIABLE))
            else:
                txt_id = dpg.add_text(source=tag)
                self.widgets[name]["variables"][variable].append((txt_id, WidgetType.TEXT))

    def add_variable(self, cls: Mod, variable: str, getter: VariableProtocol):
        name = cls.__class__.__name__
        tag = f"{name}.{variable}"
        value = getattr(cls, variable)
        with dpg.value_registry():
            # If there is no setter, the value type has to be a string
            if not getter._has_setter:
                dpg.add_string_value(tag=tag, default_value=str(value))
            else:
                if getter._variable_type == VariableType.INTEGER:
                    dpg.add_int_value(tag=tag, default_value=value)
                elif getter._variable_type == VariableType.STRING:
                    dpg.add_string_value(tag=tag, default_value=value)
                elif getter._variable_type == VariableType.FLOAT:
                    dpg.add_double_value(tag=tag, default_value=value)
                elif getter._variable_type == VariableType.BOOLEAN:
                    dpg.add_bool_value(tag=tag, default_value=value)
        self.add_variable_gui_elements(cls, variable, getter)
        if name not in self.tracking_variables:
            self.tracking_variables[name] = []
        self.tracking_variables[name].append(
            (
                tag,
                cls,
                variable,
                getter._variable_type == VariableType.STRING or not getter._has_setter,
            )
        )

    def remove_tab(self, cls: Mod):
        """Remove the tab associated with the provided class."""
        name = cls.__class__.__name__
        self.tabs.pop(dpg.get_alias_id(name))
        dpg.delete_item(name)

    def run(self):
        try:
            dpg.show_viewport()
            dpg.set_primary_window(WINDOW_TITLE, True)
            self.hwnd = win32gui.FindWindow(None, WINDOW_TITLE)
            self._current_tab = list(self.tabs.values())[0]
            while dpg.is_dearpygui_running():
                # For each tracking variable, update the value.
                for vars in self.tracking_variables.get(self._current_tab, []):
                    try:
                        tag, cls, var, is_str = vars
                        if is_str:
                            dpg.set_value(tag, str(getattr(cls, var)))
                        else:
                            dpg.set_value(tag, getattr(cls, var))
                    except Exception:
                        # If we can't set the value, don't crash the whole program.
                        pass
                dpg.render_dearpygui_frame()
            dpg.destroy_context()
        except Exception:
            logger.exception("Unable to create GUI window!")

    def exit(self):
        dpg.stop_dearpygui()



import ctypes
from typing import Annotated

from pymhf.core.hooking import Structure
from pymhf.core.memutils import get_addressof, map_struct
from pymhf.utils.partial_struct import partial_struct


def test_new_empty_struct():
    # Test creating a new empty struct.
    @partial_struct
    class Test(Structure):
        _total_size_ = 0x40
        a: Annotated[ctypes.c_uint32, 0x0]
        b: Annotated[ctypes.c_uint32, 0x10]
        c: Annotated[ctypes.c_uint32, 0x20]
        d: Annotated[ctypes.c_uint32, 0x30]

    empty_obj = Test.new_empty()

    assert empty_obj.a == 0
    assert empty_obj.b == 0
    assert empty_obj.c == 0
    assert empty_obj.d == 0

    obj_addr = get_addressof(empty_obj)

    # Map the address to another instance of the object, change the values of that one and then check that the
    # values are shared by both instances.

    new_obj = map_struct(obj_addr, Test)
    new_obj.a = 1
    new_obj.b = 2
    new_obj.c = 3
    new_obj.d = 4

    assert empty_obj.a == 1
    assert empty_obj.b == 2
    assert empty_obj.c == 3
    assert empty_obj.d == 4



# A collection of functions which will cache things.

import json
import os
import os.path as op
from logging import getLogger
from typing import Optional

from pymem.ressources.structure import MODULEINFO

import pymhf.core._internal as _internal

logger = getLogger(__name__)


# "handle-module" cache to avoid having to get the handles and such every time we need to do a look up.
hm_cache: dict[str, tuple[int, MODULEINFO]] = {}


class OffsetCache:
    """A simple cache to store offsets once they have been found within a particular binary.
    This cached data is only correct for a specific exe unique by the hash."""

    def __init__(self):
        self._lookup: dict[str, dict[str, int]] = {}
        self.loaded = False

    @property
    def path(self) -> str:
        return op.join(_internal.CACHE_DIR, f"{_internal.BINARY_HASH}.json")

    def load(self):
        """Load the data."""
        logger.debug(f"loading cache {self.path}")
        if op.exists(self.path):
            with open(self.path, "r") as f:
                self._lookup = json.load(f)
                self.loaded = True

    def save(self):
        """Persist the cache to disk."""
        if not op.exists(op.dirname(self.path)):
            os.makedirs(op.dirname(self.path), exist_ok=True)
        with open(self.path, "w") as f:
            json.dump(self._lookup, f, indent=1)

    def get(self, pattern: str, binary: Optional[str] = None) -> Optional[int]:
        """Get the offset based on the pattern provided."""
        return self._lookup.get(binary or _internal.EXE_NAME, {}).get(pattern)

    def set(self, pattern: str, offset: int, binary: Optional[str] = None, save: bool = True):
        """Set the pattern with the given value and optionally save."""
        _binary = binary
        if _binary is None:
            _binary = _internal.EXE_NAME
        if _binary not in self._lookup:
            self._lookup[_binary] = {}
        self._lookup[_binary][pattern] = offset
        if save:
            self.save()

    def items(self, binary: Optional[str] = None):
        for pattern, offset in self._lookup.get(binary or _internal.EXE_NAME, {}).items():
            yield pattern, offset


module_map: dict[str, MODULEINFO] = {}
offset_cache = OffsetCache()



# Test functions in the pymhf.core.importing file
import pytest

from pymhf.core.importing import parse_file_for_mod

# "files" which will pass.
STANDARD_IMPORT = """
import pymhf
class Thing(pymhf.Mod):
    pass
"""

STANDARD_FROM_IMPORT = """
from pymhf import Mod
class Thing(Mod):
    pass
"""

ALIAS_IMPORT = """
import pymhf as pmf
class Thing(pmf.Mod):
    pass
"""

ALIAS_FROM_IMPORT = """
from pymhf import Mod as mod
class Thing(mod):
    pass
"""

FULL_PATH_STANDARD_IMPORT = """
import pymhf.core.mod_loader
class Thing(pymhf.core.mod_loader.Mod):
    pass
"""

FULL_PATH_STANDARD_FROM_IMPORT = """
from pymhf.core.mod_loader import Mod
class Thing(Mod):
    pass
"""

FULL_PATH_ALIAS_IMPORT = """
import pymhf.core.mod_loader as pmf
class Thing(pmf.Mod):
    pass
"""

FULL_PATH_ALIAS_FROM_IMPORT = """
from pymhf.core.mod_loader import Mod as mod
class Thing(mod):
    pass
"""

# "files" which will fail.
NO_IMPORT = """
class Thing(Mod):
    pass
"""

NO_IMPORT2 = """
class Thing(pymhf.Mod):
    pass
"""

INCORRECT_IMPORT = """
from pymhf.core import Mod
class Thing(Mod):
    pass
"""

NO_MOD_CLASS = """
from pymhf import Mod
class Thing():
    pass
"""


@pytest.mark.parametrize(
    "data,result",
    [
        (STANDARD_IMPORT, True),
        (STANDARD_FROM_IMPORT, True),
        (ALIAS_IMPORT, True),
        (ALIAS_FROM_IMPORT, True),
        (FULL_PATH_STANDARD_IMPORT, True),
        (FULL_PATH_STANDARD_FROM_IMPORT, True),
        (FULL_PATH_ALIAS_IMPORT, True),
        (FULL_PATH_ALIAS_FROM_IMPORT, True),
        (NO_IMPORT, False),
        (NO_IMPORT2, False),
        (INCORRECT_IMPORT, False),
        (NO_MOD_CLASS, False),
    ],
)
def test_parse_file_for_mod(data: str, result: bool):
    assert parse_file_for_mod(data) is result



import logging
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor
from ctypes import byref, c_ulong, create_unicode_buffer, windll
from functools import wraps
from typing import Optional

import psutil
import pywinctl as pwc
import win32gui
import win32process

import pymhf.core._internal as _internal

logger = logging.getLogger(__name__)


def get_main_window_handle() -> Optional[int]:
    """Return the handle of the main running application window if possible.
    This will correspond to the HWND for the window belonging to the PID of the main running process.
    """
    windows = {x.getHandle(): x for x in pwc.getAllWindows()}
    main_pid_hwnds = get_hwnds_for_pid(_internal.PID)
    wins = [x for x, y in windows.items() if (x in main_pid_hwnds and y.title != "pyMHF")]
    if len(wins) == 0:
        logger.warning(f"Cannot find window handle for PID {_internal.PID}")
        return None
    elif len(wins) > 1:
        logger.error(
            f"Found multiple windows for PID {_internal.PID}: {main_pid_hwnds}.\n"
            "Picking the first arbitrarily but this may not be correct."
        )
        return wins[0]
    else:
        return wins[0]


def get_hwnds_for_pid(pid: int) -> list[int]:
    """Return all HWND's for the provided PID."""

    def callback(hwnd: int, hwnds: list[int]):
        _, found_pid = win32process.GetWindowThreadProcessId(hwnd)

        if found_pid == pid:
            hwnds.append(hwnd)
        return True

    hwnds = []
    win32gui.EnumWindows(callback, hwnds)
    return hwnds


def get_window_by_handle(handle: int) -> Optional[pwc.Window]:
    windows = {x.getHandle(): x for x in pwc.getAllWindows()}
    return windows.get(handle)


def set_main_window_active(callback: Optional[Callable[[], None]] = None):
    """Set the main window as active.
    If a callback is provided, it will be called after activating the window.
    This callback must not take any arguments and any return value will be ignored.
    """
    # Make sure that we have the MAIN_HWND in case it wasn't found earlier.
    if not _internal.MAIN_HWND:
        _internal.MAIN_HWND = get_main_window_handle()
        if not _internal.MAIN_HWND:
            logger.error("Cannot set main window active as we can't find it...")
            return
    if not is_main_window_foreground():
        if main_window := get_window_by_handle(_internal.MAIN_HWND):
            main_window.activate()
            if callback is not None:
                callback()


def is_main_window_foreground() -> bool:
    return win32gui.GetForegroundWindow() == _internal.MAIN_HWND


def get_main_window() -> Optional[pwc.Window]:
    if _internal.MAIN_HWND is not None:
        main_window = get_window_by_handle(_internal.MAIN_HWND)
        return main_window


# def dump_resource(res, fname):
#     with open(op.join(_internal.CWD, fname), "w") as f:
#         f.write(json.dumps(res, indent=2))


def safe_assign_enum(enum, index: int):
    """Safely try and get the enum with the associated integer value.
    If the index isn't one in the enum return the original index.
    """
    try:
        return enum(index)
    except ValueError:
        return index


def get_foreground_window_title() -> Optional[str]:
    hWnd = windll.user32.GetForegroundWindow()
    length = windll.user32.GetWindowTextLengthW(hWnd)
    buf = create_unicode_buffer(length + 1)
    windll.user32.GetWindowTextW(hWnd, buf, length + 1)

    # 1-liner alternative: return buf.value if buf.value else None
    if buf.value:
        return buf.value
    else:
        return None


def get_foreground_pid():
    handle = windll.user32.GetForegroundWindow()
    pid = c_ulong()
    windll.user32.GetWindowThreadProcessId(handle, byref(pid))
    return pid.value


def does_pid_have_focus(pid: int) -> bool:
    return pid == get_foreground_pid()


# TODO: Do something about this...
# class AutosavingConfig(ConfigParser):
#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         self._filename: str
#         self._encoding: Optional[str]

#     def read(self, filenames, encoding=None):
#         super().read(filenames, encoding)
#         self._filename = filenames
#         self._encoding = encoding

#     def set(self, section: str, option: str, value=None):
#         if value is not None:
#             val = str(value)
#         else:
#             val = value
#         try:
#             super().set(section, option, val)
#             with open(self._filename, "w", encoding=self._encoding) as f:
#                 self.write(f, space_around_delimiters=True)
#         except Exception:
#             logger.exception("Error saving file")


def saferun(func, *args, **kwargs):
    """Safely run the specified function with args and kwargs.

    Any exception raised will be shown and ignored
    """
    ret = None
    try:
        ret = func(*args, **kwargs)
    except Exception:
        logger.exception(f"There was an exception while calling {func}:")
    return ret


def saferun_decorator(func: Callable):
    """Safely run the decorated function so that any errors are caught and logged."""

    @wraps(func)
    def inner(*args, **kwargs):
        ret = None
        try:
            ret = func(*args, **kwargs)
        except Exception:
            logger.exception(f"There was an exception while calling {func}:")
        return ret

    return inner


# Some experimental functions. Private until they can be made to work...


def _pause_or_resume_process(pid: int, pause: bool):
    proc = psutil.Process(pid)
    if pause:
        proc.suspend()
    else:
        proc.resume()


def _pause_process(pid: Optional[int] = None):
    if not pid:
        pid = _internal.PID
    logger.info(f"Pausing process {pid}")
    with ThreadPoolExecutor(max_workers=1) as exc:
        exc.submit(_pause_or_resume_process, pid, True)


def _resume_process(pid: Optional[int] = None):
    if not pid:
        pid = _internal.PID
    logger.info(f"Resuming process {pid}")
    with ThreadPoolExecutor(max_workers=1) as exc:
        exc.submit(_pause_or_resume_process, pid, False)



# A few functions to interface with iced_x86.

import logging
import struct

try:
    from iced_x86 import (
        BlockEncoder,
        Code,
        Decoder,
        FastFormatter,
        Instruction,
        MemoryOperand,
        Mnemonic,
        Register,
    )

    HAS_ICED = True
except ImportError:
    HAS_ICED = False

logger = logging.getLogger(__name__)


BITS = struct.calcsize("P") * 8


def create_jmp_bytes(target: int, rip: int):
    """Assemble the required bytes to jump to some address."""
    instructions = []
    instructions.append(Instruction.create_branch(Code.JMP_REL32_64, target))
    encoder = BlockEncoder(64)
    encoder.add_many(instructions)
    return encoder.encode(rip)


def generate_load_stack_pointer_bytes(buff_addr: int, rip: int, bits: int = 64) -> bytes:
    if bits == 64:
        return load_rsp(buff_addr, rip)
    elif bits == 32:
        return load_esp(buff_addr, rip)
    else:
        raise ValueError("Number of bits must be 32 or 64")


def load_rsp(buff_addr: int, rip: int) -> bytes:
    """Assemble the required bytes to write the value of the rsp register into a buffer which can be accessed
    by the detour. This is for getting the caller address in a 64 bit process.
    The asm which is assembled is
    ```x86asm
    mov rax, [rsp]
    mov [rsp_buff_addr], rax
    ```
    """
    instructions = [
        Instruction.create_reg_mem(
            Code.MOV_R64_RM64,
            Register.RAX,
            MemoryOperand(Register.RSP),
        ),
        Instruction.create_mem_reg(
            Code.MOV_MOFFS64_RAX,
            MemoryOperand(displ=buff_addr, displ_size=8),
            Register.RAX,
        ),
    ]
    encoder = BlockEncoder(64)
    encoder.add_many(instructions)
    return encoder.encode(rip)


def load_esp(buff_addr: int, rip: int) -> bytes:
    """Assemble the required bytes to write the value of the esp register into a buffer which can be accessed
    by the detour. This is for getting the caller address in a 32 bit process.
    The asm which is assembled is
    ```x86asm
    mov eax, [esp]
    mov [rsp_buff_addr], eax
    ```
    """
    instructions = [
        Instruction.create_reg_mem(
            Code.MOV_R32_RM32,
            Register.EAX,
            MemoryOperand(Register.ESP),
        ),
        Instruction.create_mem_reg(
            Code.MOV_MOFFS32_EAX,
            MemoryOperand(displ=buff_addr, displ_size=4),
            Register.EAX,
        ),
    ]
    encoder = BlockEncoder(32)
    encoder.add_many(instructions)
    return encoder.encode(rip)


def get_first_jmp_addr(data: bytes, ip: int) -> int:
    """Get the address of the first jmp instruction found in the bytes when disassembled."""
    decoder = Decoder(BITS, data, ip=ip)
    for instr in decoder:
        if instr.mnemonic == Mnemonic.JMP:
            return instr.near_branch_target
    return 0


def disassemble(data: bytes, ip: int) -> None:
    """Utility function for disassembly bytes."""
    formatter = FastFormatter()
    decoder = Decoder(BITS, data, ip=ip)
    for instruction in decoder:
        disasm = formatter.format(instruction)
        logger.info(f"{instruction.ip:016X} {disasm}")
    logger.info("")



from typing import Optional, Union

from pymhf.core._types import FUNCDEF

# Load order:
# 1. Look up offsets based on binary hash in offset cache (#TODO: implement)
# 2. Look up patterns
# 3. Use offsets

# NOTE: The above is just temporary. Need to figure out a more robust way to specify this.
# (Well, the order of 2 and 3 should be swapped likely. Need a way to specify which to use...)


class ModuleData:
    FUNC_OFFSETS: dict[str, Union[int, dict[str, int]]] = {}
    FUNC_PATTERNS: dict[str, Union[str, dict[str, str]]] = {}
    FUNC_CALL_SIGS: dict[str, Union[FUNCDEF, dict[str, FUNCDEF]]] = {}
    FUNC_BINARY: Optional[str] = None


# TODO: Remove.
module_data = ModuleData()



import ctypes
import ctypes.wintypes as wintypes
import hashlib
import os
from io import BufferedReader

import psutil
import pymem
from pymem.ressources.structure import (
    MEMORY_BASIC_INFORMATION,
    MEMORY_BASIC_INFORMATION32,
    MEMORY_BASIC_INFORMATION64,
    MEMORY_PROTECTION,
    MEMORY_STATE,
    MEMORY_TYPES,
    MODULEINFO,
    SYSTEM_INFO,
)
from typing_extensions import Union, cast

from pymhf.utils.winapi import (
    IMAGE_DOS_HEADER,
    IMAGE_DOS_SIGNATURE,
    IMAGE_FILE_HEADER,
    IMAGE_NT_SIGNATURE,
    IMAGE_SCN_MEM_EXECUTE,
    IMAGE_SCN_MEM_WRITE,
    IMAGE_SECTION_HEADER,
    GetSystemInfo,
    VirtualQueryEx,
)


def _is_hashable_page(mbi: Union[MEMORY_BASIC_INFORMATION32, MEMORY_BASIC_INFORMATION64]) -> bool:
    """Check if a memory page is suitable for hashing. The page must not change during runtime and/or
    between runs."""
    if mbi.State != MEMORY_STATE.MEM_COMMIT:
        return False
    if mbi.Type != MEMORY_TYPES.MEM_IMAGE:
        return False
    if mbi.Protect & (
        MEMORY_PROTECTION.PAGE_GUARD
        | MEMORY_PROTECTION.PAGE_WRITECOPY
        | MEMORY_PROTECTION.PAGE_EXECUTE_WRITECOPY
    ):
        return False
    if mbi.Protect & (MEMORY_PROTECTION.PAGE_READWRITE | MEMORY_PROTECTION.PAGE_EXECUTE_READWRITE):
        return False
    if not (mbi.Protect & (MEMORY_PROTECTION.PAGE_EXECUTE | MEMORY_PROTECTION.PAGE_EXECUTE_READ)):
        return False
    return True


def _get_page_size() -> int:
    """Get the system page size. Defaults to 4096 if it cannot be determined."""
    sys_info = SYSTEM_INFO()
    GetSystemInfo(ctypes.byref(sys_info))
    return sys_info.dwPageSize or 4096


def _get_main_module(pm_binary: pymem.Pymem) -> MODULEINFO:
    binary_path = psutil.Process(pm_binary.process_id).exe().lower()
    binary_exe = os.path.basename(binary_path).lower()

    main_module = None
    modules = list(pm_binary.list_modules())
    for module in modules:
        if module.filename.lower() == binary_path or module.name.lower() == binary_exe:
            main_module = module
            break
    if not main_module:
        main_module = modules[0]  # Usually the first module is the main module
        # Maybe raising an error or returning `None` here instead would be safer
        # raise OSError(f"Could not find main module for process {pid}")

    return main_module


def _get_sections_info(pm_binary: pymem.Pymem, address: int) -> tuple[int, int]:
    """Get the base address and number of sections in the PE file at the given address."""
    dos_header = cast(IMAGE_DOS_HEADER, pm_binary.read_ctype(address, IMAGE_DOS_HEADER()))
    if dos_header.e_magic != IMAGE_DOS_SIGNATURE:
        raise ValueError(f"Invalid DOS header magic for address 0x{address:X}")

    address += dos_header.e_lfanew
    signature = pm_binary.read_ctype(address, wintypes.DWORD())
    if signature != IMAGE_NT_SIGNATURE:
        raise ValueError(f"Invalid PE header signature for address 0x{address:X}")

    address += ctypes.sizeof(wintypes.DWORD)
    file_header = cast(IMAGE_FILE_HEADER, pm_binary.read_ctype(address, IMAGE_FILE_HEADER()))

    num_sections = int(file_header.NumberOfSections)
    opt_header_size = int(file_header.SizeOfOptionalHeader)
    sections_base = address + ctypes.sizeof(IMAGE_FILE_HEADER) + opt_header_size

    return sections_base, num_sections


def _get_read_only_sections(
    pm_binary: pymem.Pymem,
    sections_base: int,
    num_sections: int,
    max_module_size: int,
):
    """Get a list of read-only sections in the PE file at the given address."""
    sections = []
    for i in range(num_sections):
        section_address = sections_base + i * ctypes.sizeof(IMAGE_SECTION_HEADER)
        section_header = cast(
            IMAGE_SECTION_HEADER, pm_binary.read_ctype(section_address, IMAGE_SECTION_HEADER())
        )

        characteristics = section_header.Characteristics
        if not (characteristics & IMAGE_SCN_MEM_EXECUTE) or (characteristics & IMAGE_SCN_MEM_WRITE):
            continue

        virtual_addr = int(section_header.VirtualAddress)
        virtual_size = int(section_header.Misc.VirtualSize) or int(section_header.SizeOfRawData)
        if virtual_addr == 0 or virtual_size == 0:
            continue

        end_addr = min(virtual_addr + virtual_size, max_module_size)
        if end_addr <= virtual_addr:
            continue

        section = (
            virtual_addr,
            end_addr - virtual_addr,
            bytes(bytearray(section_header.Name)).rstrip(b"\x00").decode(errors="ignore"),
        )
        sections.append(section)

    return sections


def hash_bytes_from_file(fileobj: BufferedReader, _bufsize: int = 2**18) -> str:
    # Essentially implement hashlib.file_digest since it's python 3.11+
    # cf. https://github.com/python/cpython/blob/main/Lib/hashlib.py#L195
    digestobj = hashlib.sha1()
    buf = bytearray(_bufsize)  # Reusable buffer to reduce allocations.
    view = memoryview(buf)
    while True:
        size = fileobj.readinto(buf)
        if size == 0:
            break  # EOF
        digestobj.update(view[:size])
    return digestobj.hexdigest()


def hash_bytes_from_memory(pm_binary: pymem.Pymem, _bufsize: int = 2**18) -> str:
    """Hash the bytes of the main module of the given `pymem.Pymem` instance.
    In order to ensure that the hash is stable across runs, this only read from sections that are not expected
    to change between runs."""
    process_handle = pm_binary.process_handle
    pid = pm_binary.process_id
    if not process_handle or not pid:
        raise ValueError("Pymem instance does not have a valid process handle")

    main_module = _get_main_module(pm_binary)
    if not main_module:
        raise OSError(f"Could not find main module for process {pid}")

    base_address = main_module.lpBaseOfDll
    module_size = main_module.SizeOfImage
    if not base_address or not module_size:
        raise OSError("Failed to resolve main module base/size")

    sections_base, num_sections = _get_sections_info(pm_binary, base_address)
    sections = _get_read_only_sections(pm_binary, sections_base, num_sections, module_size)
    if not sections:
        raise ValueError("No read-only sections found in the main module")
    sections.sort(key=lambda s: s[0])

    page_size = _get_page_size()
    digest = hashlib.sha1()
    buffer = (ctypes.c_ubyte * _bufsize)()
    for rva, size, name in sections:
        start = base_address + rva
        end = start + size
        address = start

        while address < end:
            page = MEMORY_BASIC_INFORMATION()
            if not VirtualQueryEx(
                process_handle,
                ctypes.c_void_p(address),
                ctypes.byref(page),
                ctypes.sizeof(page),
            ):
                address += page_size
                continue

            region_end = min(end, address + page.RegionSize)
            if not _is_hashable_page(page):
                address = region_end
                continue

            current = address
            while current < region_end:
                to_read = min(_bufsize, region_end - current)
                buffer = pm_binary.read_bytes(current, to_read)
                if len(buffer) == 0:
                    current = (current + page_size) & ~(page_size - 1)
                    if current < address:
                        current = address + page_size
                    continue

                digest.update(memoryview(buffer)[: len(buffer)])
                current += len(buffer)

            address = region_end

    return digest.hexdigest()



import asyncio
import builtins
import ctypes
import locale
import logging
import logging.handlers
import os
import os.path as op
import time
import traceback
from concurrent.futures import ThreadPoolExecutor
from functools import partial
from typing import Optional

import pymem
import pymem.process

from pymhf.core.utils import get_main_window_handle

socket_logger_loaded = False
executor = None
ready = False

try:
    import pymhf.core._internal as _internal
    from pymhf.utils.config import canonicalize_setting

    rootLogger = logging.getLogger("")

    logging_config = _internal.CONFIG.get("logging", {}) or {}

    log_level = logging_config.get("log_level", "info")
    if log_level.lower() == "debug":
        rootLogger.setLevel(logging.DEBUG)
    else:
        rootLogger.setLevel(logging.INFO)

    if not logging_config.get("shown", True):
        # In this case we just want to log to a file somewhere... For now, default to a folder called logs
        # where the module path is.
        formatter = logging.Formatter("%(asctime)s %(name)-24s %(levelname)-6s %(message)s")

        _log_dir = logging_config.get("log_dir", "{CURR_DIR}")
        if op.isdir(_internal.MODULE_PATH):
            module_dir = _internal.MODULE_PATH
        else:
            module_dir = op.join(_internal.MODULE_PATH, "..")
        log_dir = canonicalize_setting(_log_dir, None, module_dir, op.dirname(_internal.BINARY_PATH), "logs")
        # If this ends up not being able to be resolved, fallback to logs in the same directory as the module.
        if log_dir is None:
            if op.isdir(_internal.MODULE_PATH):
                log_dir = op.join(_internal.MODULE_PATH, "logs")
            else:
                log_dir = op.join(_internal.MODULE_PATH, "..", "logs")
        os.makedirs(log_dir, exist_ok=True)
        file_handler = logging.FileHandler(
            op.join(log_dir, f"pymhf-{time.strftime('%Y%m%dT%H%M%S')}.log"), encoding="utf-8"
        )
        file_handler.setFormatter(formatter)
        rootLogger.addHandler(file_handler)
    else:
        socketHandler = logging.handlers.SocketHandler("localhost", logging.handlers.DEFAULT_TCP_LOGGING_PORT)
        rootLogger.addHandler(socketHandler)
    logging.info("Loading pyMHF...")
    socket_logger_loaded = True

    from pymhf.core._types import LoadTypeEnum

    _internal.LOAD_TYPE = LoadTypeEnum(_internal.LOAD_TYPE)

    # Assign the sentinel and load it from the provided address which is allocated in this process by the
    # parent process. If the main process isn't actually waiting on this it won't ever be read, but it's
    # simpler to just allocate it anyway.
    sentinel = ctypes.c_bool(False)
    if _internal._SENTINEL_PTR:
        sentinel = ctypes.c_bool.from_address(_internal._SENTINEL_PTR)

    _module_path = _internal.MODULE_PATH
    if op.isfile(_module_path):
        _module_path = op.dirname(_module_path)
    _binary_dir = None
    if _internal.BINARY_PATH:
        _binary_dir = op.dirname(_internal.BINARY_PATH)

    internal_mod_folder = _internal.CONFIG.get("internal_mod_dir")
    internal_mod_folder = canonicalize_setting(internal_mod_folder, "pymhf", _module_path, _binary_dir)

    mod_folder = _internal.CONFIG.get("mod_dir")
    mod_folder = canonicalize_setting(mod_folder, "pymhf", _module_path, _binary_dir)

    import keyboard._winkeyboard as kwk

    # Prefill the key name tables to avoid taking a hit when hooking.
    kwk._setup_name_tables()

    import pymhf.core.caching as cache
    from pymhf.core.hooking import hook_manager
    from pymhf.core.memutils import getsize
    from pymhf.core.mod_loader import mod_manager
    from pymhf.core.protocols import (
        ESCAPE_SEQUENCE,
        READY_ACK_SEQUENCE,
        READY_ASK_SEQUENCE,
        ExecutionEndedException,
        custom_exception_handler,
    )

    try:
        from pymhf.gui.gui import GUI
    except ModuleNotFoundError:
        # If we can't import this, then DearPyGUI is missing, so we won't create the GUI.
        GUI = None
    from pymhf.utils.imports import get_imports

    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    # Since we are running inside a thread, `asyncio.get_event_loop` will
    # generally fail.
    # Detect this and create a new event loop anyway since we are running in a
    # thread under the process we have been injected into, and not the original
    # python thread that is running the show.
    try:
        loop = asyncio.get_event_loop()
    except (RuntimeError, ValueError):
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    # Set the custom exception handler on the loop
    loop.set_exception_handler(custom_exception_handler)

    # NOTE: This class MUST be defined here. If it's defined in a separate file
    # then the hack done to persist data to the current global context will not
    # work.
    class ExecutingProtocol(asyncio.Protocol):
        """A protocol factory to be passed to a asyncio loop.create_server call
        which will accept requests, execute them and persist any variables to
        globals().
        """

        def connection_made(self, transport: asyncio.transports.WriteTransport):
            self.transport: asyncio.transports.WriteTransport = transport
            # peername = transport.get_extra_info('peername')
            # self.write(f'Connection from {peername} ')
            # Overwrite print so that any `print` statements called in the commands
            # to be executed will be written back out of the socket they came in.
            globals()["print"] = partial(builtins.print, file=self)

        def write(self, value: str):
            """
            Method to allow this protocol to be used as a file to write to.

            This allows us to have `print` write to this protocol.
            """
            self.transport.write(value.encode())

        def data_received(self, __data: bytes):
            # Have an "escape sequence" which will force this to exit.
            # This way we can kill it if need be from the other end.
            if __data == ESCAPE_SEQUENCE:
                print("\nReceived exit command!")
                raise ExecutionEndedException
            elif __data == READY_ASK_SEQUENCE:
                print("\nReceived ready ask command")
                print(f"Are we ready? {ready}")
                self.transport.write(READY_ACK_SEQUENCE)
                return
            try:
                exec(__data.decode())
            except Exception:
                print(traceback.format_exc())
            else:
                self.persist_to_globals(locals())

        def persist_to_globals(self, data: dict):
            """Take the dict which was determined by calling `locals()`, and update `gloabsl()` with it."""
            data.pop("self")
            data.pop(f"_{type(self).__name__}__data")
            globals().update(data)

        def eof_received(self):
            # Do nothing.
            pass

        def connection_lost(self, exc):
            # Once the connection is lost. Restore `print` back to normal.
            globals()["print"] = builtins.print

    def top_globals(limit: Optional[int] = 10):
        """Return the top N objects in globals() by size (in bytes)."""
        globs = globals()
        data = []
        for key, value in globs.items():
            if not key.startswith("__"):
                try:
                    data.append((key, *getsize(value)))
                except TypeError:
                    pass
        data.sort(key=lambda x: x[1], reverse=True)
        if limit is not None:
            return data[:limit]
        else:
            return data

    # Patch the locale to make towupper work.
    # Python must change this so we change it back otherwise calls to `towupper`
    # in the various functions to set and get keypresses don't work correctly.
    locale.setlocale(locale.LC_CTYPE, "C")

    executor = ThreadPoolExecutor(2, thread_name_prefix="pyMHF_Internal_Executor")

    binary = pymem.Pymem(_internal.EXE_NAME, exact_match=True)
    cache.module_map = {x.name: x for x in pymem.process.enum_process_module(binary.process_handle)}

    # Read the imports
    if _internal.BINARY_PATH:
        _internal.imports = get_imports(_internal.BINARY_PATH)
    for fpath in _internal.INCLUDED_ASSEMBLIES.values():
        _internal.imports.update(get_imports(fpath))

    # Load the offset cache.
    if not cache.offset_cache.loaded:
        cache.offset_cache.load()

    mod_manager.hook_manager = hook_manager
    # First, load our internal mod before anything else.
    if internal_mod_folder is not None:
        logging.debug(f"Loading internal mods: {internal_mod_folder}")
        # If the mod folder isn't absolute, assume it's relative to the library directory.
        if not op.isabs(internal_mod_folder):
            internal_mod_folder = op.join(_internal.MODULE_PATH, internal_mod_folder)
        if op.exists(internal_mod_folder):
            mod_manager.load_mod_folder(internal_mod_folder, bind=False)
        else:
            logging.warning(
                f"Cannot load internal mod directory: {internal_mod_folder}. "
                "Please make sure it exists or that the path is correct in the pymhf.toml file."
            )

    logging.info("pyMHF injection complete!")

    # Also load any mods after all the internal hooks:
    start_time = time.time()
    bold = "\u001b[4m"
    reset = "\u001b[0m"
    logging.info(bold + "Loading mods" + reset)
    _loaded_mods = 0
    _loaded_hooks = 0
    try:
        if _internal.LOAD_TYPE == LoadTypeEnum.SINGLE_FILE:
            # For a single file mod, we just load that file.
            _loaded_mods, _loaded_hooks = mod_manager.load_single_mod(_internal.MODULE_PATH)
        elif _internal.LOAD_TYPE == LoadTypeEnum.MOD_FOLDER:
            _loaded_mods, _loaded_hooks = mod_manager.load_mod_folder(_internal.MODULE_PATH, deep_search=True)
        else:  # Loading a library.
            if mod_folder is not None:
                _loaded_mods, _loaded_hooks = mod_manager.load_mod_folder(mod_folder, deep_search=True)
            else:
                logging.warning(
                    """You have not configured the `mod_dir` variable in the pymhf.toml file.
                    Please do so so that you can load mods."""
                )
    except Exception:
        logging.error(traceback.format_exc())
    _mods_str = "mod"
    if _loaded_mods != 1:
        _mods_str = "mods"
    _hooks_str = "hook"
    if _loaded_hooks != 1:
        _hooks_str = "hooks"
    logging.info(
        f"Loaded {_loaded_mods} {_mods_str} and {_loaded_hooks} {_hooks_str} in "
        f"{time.time() - start_time:.3f}s"
    )

    mod_manager._assign_mod_instances()

    _internal.MAIN_HWND = get_main_window_handle()

    for func_name, hook_class in hook_manager.failed_hooks.items():
        offset = hook_class.target
        _data = (ctypes.c_char * 0x20).from_address(offset)
        rootLogger.error(f"Hook {func_name} first 0x20 bytes: {_data.value.hex()}")

    # Each client connection will create a new protocol instance
    coro = loop.create_server(ExecutingProtocol, "127.0.0.1", 6770)
    server = loop.run_until_complete(coro)

    futures = []
    if _internal.CONFIG.get("gui", {}).get("shown", True) and GUI is not None:
        gui = GUI(mod_manager, _internal.CONFIG)
        # For each mod, add the corresponding tab to the gui.
        for mod in mod_manager.mods.values():
            gui.add_tab(mod)
        # Add the settings tab so that we may configure various settings.
        gui.add_hex_tab()
        gui.add_settings_tab()
        gui.add_details_tab()

        # TODO: This needs to have some exception handling because if something
        # goes wrong in here it will just fail "silently".
        futures.append(executor.submit(gui.run))

    logging.info(f"Serving on executor {server.sockets[0].getsockname()}")

    # Finally, before we run forever, set the sentinel value to True so that if the main process was waiting
    # for the injected code to complete before starting the process it can now resume it.
    sentinel.value = True

    loop.run_forever()

    # Close the server.
    server.close()
    loop.run_until_complete(server.wait_closed())
    loop.close()

except Exception:
    # If we hit this, something has gone wrong. Log to the current directory.

    try:
        # Try and log to the current working directory.
        # Sometimes this may fail as the error is an "internal" one, so we will
        # add a fail-safe to log to the users' home directory so that it at
        # least is logged somewhere.
        # TODO: Document this behaviour.
        import pymhf.core._internal as _internal

        with open(op.join(_internal.CWD, "CRITICAL_ERROR.txt"), "w") as f:
            traceback.print_exc(file=f)
            if socket_logger_loaded:
                logging.error("An error occurred while loading pymhf:")
                logging.error(traceback.format_exc())
    except Exception:
        with open(op.join(op.expanduser("~"), "CRITICAL_ERROR.txt"), "w") as f:
            traceback.print_exc(file=f)
finally:
    if executor is not None:
        executor.shutdown(wait=False, cancel_futures=True)



import os.path as op
import tempfile

from pymhf.utils.config import canonicalize_setting

CWD = op.dirname(__file__)


def test_canonicalize_setting():
    # Create log directory in the current directory.
    assert canonicalize_setting(".", None, CWD, CWD, "logs") == op.join(CWD, "logs")
    assert canonicalize_setting("{CURR_DIR}", None, CWD, CWD, "logs") == op.join(CWD, "logs")

    with tempfile.TemporaryDirectory() as temp_dir:
        assert canonicalize_setting(temp_dir, None, CWD, CWD, "logs") == temp_dir
